"""
Pipeline modules for stacking and subsampling of images.
"""

import time
import warnings

from typing import List, Optional, Tuple

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import progress, memory_frames, stack_angles, angle_average
from pynpoint.util.image import rotate_images


class StackAndSubsetModule(ProcessingModule):
    """
    Pipeline module for stacking subsets of images and/or selecting a random sample of images.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 random: Optional[int] = None,
                 stacking: Optional[int] = None,
                 combine: str = 'mean',
                 max_rotation: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. Should be different from
            *image_in_tag*.
        random : int, None
            Number of random images. All images are used if set to None.
        stacking : int, None
            Number of stacked images per subset. No stacking is applied if set to None.
        combine : str
            Method for combining images ('mean' or 'median'). The angles are always mean-combined.
        max_rotation : float, None
            Maximum allowed field rotation (deg) throughout each subset of stacked images when
            `stacking` is not None. No restriction on the field rotation is applied if set to
            None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_random = random
        self.m_stacking = stacking
        self.m_combine = combine
        self.m_max_rotation = max_rotation

        if self.m_stacking is None and self.m_random is None:
            warnings.warn('Both \'stacking\' and \'random\' are set to None.')

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Stacks subsets of images and/or selects a random subset. Also
        the parallactic angles are mean-combined if images are stacked.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _stack_subsets(nimages: int,
                           im_shape: Tuple[int, ...],
                           parang: np.ndarray) -> Tuple[Tuple[int, ...], np.ndarray, np.ndarray]:

            im_new = None
            parang_new = None

            if self.m_stacking is not None:
                if self.m_max_rotation is not None:
                    frames = stack_angles(self.m_stacking, parang, self.m_max_rotation)
                else:
                    frames = memory_frames(self.m_stacking, nimages)

                nimages_new = np.size(frames)-1

                if parang is None:
                    parang_new = None
                else:
                    parang_new = np.zeros(nimages_new)

                im_new = np.zeros((nimages_new, im_shape[1], im_shape[2]))

                start_time = time.time()

                for i in range(nimages_new):
                    progress(i, nimages_new, 'Stacking subsets of images...', start_time)

                    if parang is not None:
                        # parang_new[i] = np.mean(parang[frames[i]:frames[i+1]])
                        parang_new[i] = angle_average(parang[frames[i]:frames[i+1]])

                    im_subset = self.m_image_in_port[frames[i]:frames[i+1], ]

                    if self.m_combine == 'mean':
                        im_new[i, ] = np.mean(im_subset, axis=0)
                    elif self.m_combine == 'median':
                        im_new[i, ] = np.median(im_subset, axis=0)

                im_shape = im_new.shape

            else:
                if parang is not None:
                    parang_new = np.copy(parang)

            return im_shape, im_new, parang_new

        @typechecked
        def _random_subset(im_shape: Tuple[int, ...],
                           im_new: np.ndarray,
                           parang_new: np.ndarray) -> Tuple[int, np.ndarray, np.ndarray]:

            if self.m_random is not None:
                choice = np.random.choice(im_shape[0], self.m_random, replace=False)
                choice = list(np.sort(choice))

                if parang_new is None:
                    parang_new = None
                else:
                    parang_new = parang_new[choice]

                if self.m_stacking is None:
                    im_new = self.m_image_in_port[list(choice), ]
                else:
                    im_new = im_new[choice, ]

            if self.m_random is None and self.m_stacking is None:
                nimages = 0
            elif im_new.ndim == 2:
                nimages = 1
            elif im_new.ndim == 3:
                nimages = im_new.shape[0]

            return nimages, im_new, parang_new

        non_static = self.m_image_in_port.get_all_non_static_attributes()

        im_shape = self.m_image_in_port.get_shape()
        nimages = im_shape[0]

        if self.m_random is not None:
            if self.m_stacking is None and im_shape[0] < self.m_random:
                raise ValueError('The number of images of the destination subset is larger than '
                                 'the number of images in the source.')

            if self.m_stacking is not None and \
                    int(float(im_shape[0])/float(self.m_stacking)) < self.m_random:
                raise ValueError('The number of images of the destination subset is larger than '
                                 'the number of images in the stacked source.')

        if 'PARANG' in non_static:
            parang = self.m_image_in_port.get_attribute('PARANG')
        else:
            parang = None

        im_shape, im_new, parang_new = _stack_subsets(nimages, im_shape, parang)
        nimages, im_new, parang_new = _random_subset(im_shape, im_new, parang_new)

        if self.m_random or self.m_stacking:
            self.m_image_out_port.set_all(im_new, keep_attributes=True)
            self.m_image_out_port.copy_attributes(self.m_image_in_port)
            self.m_image_out_port.add_attribute('INDEX', np.arange(0, nimages, 1), static=False)

            if parang_new is not None:
                self.m_image_out_port.add_attribute('PARANG', parang_new, static=False)

            if 'NFRAMES' in non_static:
                self.m_image_out_port.del_attribute('NFRAMES')

            history = f'stacking = {self.m_stacking}, random = {self.m_random}'
            self.m_image_out_port.add_history('StackAndSubsetModule', history)

        self.m_image_out_port.close_port()


class StackCubesModule(ProcessingModule):
    """
    Pipeline module for calculating the mean or median of each original data cube associated with a
    database tag.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 combine: str = 'mean') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry with the mean or median collapsed images that are written
            as output. Should be different from *image_in_tag*.
        combine : str
            Method to combine the images ('mean' or 'median').

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_combine = combine

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Uses the ``NFRAMES`` attribute to select the images of each cube,
        calculates the mean or median of each cube, and saves the data and attributes.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_in_port.tag == self.m_image_out_port.tag:
            raise ValueError('Input and output port should have a different tag.')

        non_static = self.m_image_in_port.get_all_non_static_attributes()
        nframes = self.m_image_in_port.get_attribute('NFRAMES')

        if 'PARANG' in non_static:
            parang = self.m_image_in_port.get_attribute('PARANG')
        else:
            parang = None

        current = 0
        parang_new = []

        start_time = time.time()
        for i, frames in enumerate(nframes):
            progress(i, len(nframes), 'Stacking images per FITS cube...', start_time)

            if self.m_combine == 'mean':
                im_stack = np.mean(self.m_image_in_port[current:current+frames, ], axis=0)
            elif self.m_combine == 'median':
                im_stack = np.median(self.m_image_in_port[current:current+frames, ], axis=0)

            self.m_image_out_port.append(im_stack, data_dim=3)

            if parang is not None:
                parang_new.append(np.mean(parang[current:current+frames]))

            current += frames

        nimages = np.size(nframes)

        self.m_image_out_port.copy_attributes(self.m_image_in_port)

        if 'INDEX' in non_static:
            index = np.arange(0, nimages, 1, dtype=int)
            self.m_image_out_port.add_attribute('INDEX', index, static=False)

        if 'NFRAMES' in non_static:
            nframes = np.ones(nimages, dtype=int)
            self.m_image_out_port.add_attribute('NFRAMES', nframes, static=False)

        if 'PARANG' in non_static:
            self.m_image_out_port.add_attribute('PARANG', parang_new, static=False)

        self.m_image_out_port.close_port()


class DerotateAndStackModule(ProcessingModule):
    """
    Pipeline module for derotating and/or stacking (i.e., taking the median or average) of the
    images, either along the time or the wavelengths dimension.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 derotate: bool = True,
                 stack: Optional[str] = None,
                 extra_rot: float = 0.,
                 dimension: str = 'time') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. The shape of the output data is
            equal to the data from ``image_in_tag``. If the argument of ``stack`` is not None,
            then the size of the collapsed dimension is equal to 1.
        derotate : bool
            Derotate the images with the ``PARANG`` attribute.
        stack : str
            Type of stacking applied after optional derotation ('mean', 'median', or None for no
            stacking).
        extra_rot : float
            Additional rotation angle of the images in clockwise direction (deg).
        dimension : str
            Dimension along which the images are stacked. Can either be 'time' or 'wavelength'. If
            the ``image_in_tag`` has three dimensions then ``dimension`` is always fixed to 'time'.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_derotate = derotate
        self.m_stack = stack
        self.m_extra_rot = extra_rot
        self.m_dimension = dimension

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Uses the ``PARANG`` attributes to derotate the images (if
        ``derotate`` is set to ``True``) and applies an optional mean or median stacking
        along the time or wavelengths dimension afterwards.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _initialize(ndim: int,
                        npix: int) -> Tuple[int, np.ndarray, Optional[np.ndarray],
                                            Optional[np.ndarray]]:

            if ndim == 2:
                nimages = 1

            elif ndim == 3:
                nimages = self.m_image_in_port.get_shape()[-3]

                if self.m_stack == 'median':
                    frames = np.array([0, nimages])

                else:
                    frames = memory_frames(memory, nimages)

            elif ndim == 4:
                nimages = self.m_image_in_port.get_shape()[-3]
                nwave = self.m_image_in_port.get_shape()[-4]

                if self.m_dimension == 'time':
                    frames = np.linspace(0, nwave, nwave+1)

                elif self.m_dimension == 'wavelength':
                    frames = np.linspace(0, nimages, nimages+1)

                else:
                    raise ValueError('The dimension should be set to \'time\' or \'wavelength\'.')

            if self.m_stack == 'mean':
                if ndim == 4:
                    if self.m_dimension == 'time':
                        im_tot = np.zeros((nwave, npix, npix))

                    elif self.m_dimension == 'wavelength':
                        im_tot = np.zeros((nimages, npix, npix))

                else:
                    im_tot = np.zeros((npix, npix))

            else:
                im_tot = None

            if self.m_stack is None and ndim == 4:
                im_none = np.zeros((nwave, nimages, npix, npix))

            else:
                im_none = None

            return nimages, frames, im_tot, im_none

        memory = self._m_config_port.get_attribute('MEMORY')

        if self.m_derotate:
            parang = self.m_image_in_port.get_attribute('PARANG')

        ndim = self.m_image_in_port.get_ndim()
        npix = self.m_image_in_port.get_shape()[-2]

        nimages, frames, im_tot, im_none = _initialize(ndim, npix)

        start_time = time.time()
        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Derotating and/or stacking images...', start_time)

            if ndim == 3:
                # Get the images and ensure they have the correct 3D shape with the following
                # three dimensions: (batch_size, height, width)
                images = self.m_image_in_port[frames[i]:frames[i+1], ]

            elif ndim == 4:
                # Process all time frames per exposure at once
                if self.m_dimension == 'time':
                    images = self.m_image_in_port[i, :, ]

                elif self.m_dimension == 'wavelength':
                    images = self.m_image_in_port[:, i, ]

            if self.m_derotate:
                if ndim == 4:
                    if self.m_dimension == 'time':
                        angles = -1.*parang + self.m_extra_rot

                    elif self.m_dimension == 'wavelength':
                        n_wavel = self.m_image_in_port.get_shape()[-4]
                        angles = np.full(n_wavel, -1.*parang[i]) + self.m_extra_rot

                else:
                    angles = -parang[frames[i]:frames[i+1]]+self.m_extra_rot

                images = rotate_images(images, angles)

            if self.m_stack is None:
                if ndim == 2:
                    self.m_image_out_port.set_all(images[np.newaxis, ...])

                elif ndim == 3:
                    self.m_image_out_port.append(images, data_dim=3)

                elif ndim == 4:
                    if self.m_dimension == 'time':
                        im_none[i] = images

                    elif self.m_dimension == 'wavelength':
                        im_none[:, i] = images

            elif self.m_stack == 'mean':
                if ndim == 4:
                    im_tot[i] = np.sum(images, axis=0)

                else:
                    im_tot += np.sum(images, axis=0)

        if self.m_stack == 'mean':
            if ndim == 4:
                im_stack = im_tot/float(im_tot.shape[0])

                if self.m_dimension == 'time':
                    self.m_image_out_port.set_all(im_stack[:, np.newaxis, ...])

                elif self.m_dimension == 'wavelength':
                    self.m_image_out_port.set_all(im_stack[np.newaxis, ...])

            else:
                im_stack = im_tot/float(nimages)
                self.m_image_out_port.set_all(im_stack[np.newaxis, ...])

        elif self.m_stack == 'median':
            if ndim == 4:
                images = self.m_image_in_port[:]

                if self.m_dimension == 'time':
                    im_stack = np.median(images, axis=1)
                    self.m_image_out_port.set_all(im_stack[:, np.newaxis, ...])

                elif self.m_dimension == 'wavelength':
                    im_stack = np.median(images, axis=0)
                    self.m_image_out_port.set_all(im_stack[np.newaxis, ...])

            else:
                im_stack = np.median(images, axis=0)
                self.m_image_out_port.set_all(im_stack[np.newaxis, ...])

        elif self.m_stack is None and ndim == 4:
            if self.m_dimension == 'time':
                self.m_image_out_port.set_all(im_none)

            elif self.m_dimension == 'wavelength':
                self.m_image_out_port.set_all(im_none)

        if self.m_derotate or self.m_stack is not None:
            self.m_image_out_port.copy_attributes(self.m_image_in_port)

        self.m_image_out_port.close_port()


class CombineTagsModule(ProcessingModule):
    """
    Pipeline module for combining tags from multiple database entries into a single tag.
    """

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tags: List[str],
                 image_out_tag: str,
                 check_attr: bool = True,
                 index_init: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tags : list(str, )
            Tags of the database entries that are read as input and combined.
        image_out_tag : str
            Tag of the database entry that is written as output. Should not be present in
            *image_in_tags*.
        check_attr : bool
            Compare non-static attributes between the tags or combine all non-static attributes
            into the new database tag.
        index_init : bool
            Reinitialize the ``INDEX`` attribute. The frames are indexed in the order of tags names
            that are provided in *image_in_tags*.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_out_port = self.add_output_port(image_out_tag)

        if image_out_tag in image_in_tags:
            raise ValueError('The \'image_out_tag\' should not be present in \'image_in_tags\'.')

        if len(image_in_tags) < 2:
            raise ValueError('The \'image_in_tags\' should contain at least two tags.')

        self.m_image_in_tags = image_in_tags
        self.m_check_attr = check_attr
        self.m_index_init = index_init

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Combines the frames of multiple tags into a single dataset
        and adds the static and non-static attributes. The values of the attributes are compared
        between the input tags to make sure that the input tags descent from the same data set.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')

        image_in_port = []
        im_shape = []

        for i, item in enumerate(self.m_image_in_tags):
            image_in_port.append(self.add_input_port(item))
            im_shape.append(image_in_port[i].get_shape()[-2:])

        if len(set(im_shape)) > 1:
            raise ValueError('The size of the images should be the same for all datasets.')

        count = 0
        start_time = time.time()
        for i, item in enumerate(self.m_image_in_tags):
            progress(i, len(self.m_image_in_tags), 'Combining datasets...', start_time)

            nimages = image_in_port[i].get_shape()[0]
            frames = memory_frames(memory, nimages)

            for j, _ in enumerate(frames[:-1]):
                im_tmp = image_in_port[i][frames[j]:frames[j+1], ]
                self.m_image_out_port.append(im_tmp)

                if self.m_index_init:
                    index = np.arange(frames[j], frames[j+1], 1) + count

                    if i == 0 and j == 0:
                        self.m_image_out_port.add_attribute('INDEX', index, static=False)
                    else:
                        for ind in index:
                            self.m_image_out_port.append_attribute_data('INDEX', ind)

            static_attr = image_in_port[i].get_all_static_attributes()
            non_static_attr = image_in_port[i].get_all_non_static_attributes()

            for key in static_attr:
                status = self.m_image_out_port.check_static_attribute(key, static_attr[key])

                if status == 1:
                    self.m_image_out_port.add_attribute(key, static_attr[key], static=True)

                elif status == -1 and key[0:7] != 'History':
                    warnings.warn(f'The static keyword {key} is already used but with a different '
                                  f'value. It is advisable to only combine tags that descend from '
                                  f'the same data set.')

            for key in non_static_attr:
                values = image_in_port[i].get_attribute(key)
                status = self.m_image_out_port.check_non_static_attribute(key, values)

                if key != 'INDEX' or (key == 'INDEX' and not self.m_index_init):

                    if self.m_check_attr:
                        if key in ('PARANG', 'STAR_POSITION', 'INDEX', 'NFRAMES'):
                            if status == 1:
                                self.m_image_out_port.add_attribute(key, values, static=False)

                            else:
                                for j in values:
                                    self.m_image_out_port.append_attribute_data(key, j)

                        else:
                            if status == 1:
                                self.m_image_out_port.add_attribute(key, values, static=False)

                            if status == -1:
                                warnings.warn(f'The non-static keyword {key} is already used but '
                                              f'with different values. It is advisable to only '
                                              f'combine tags that descend from the same data set.')

                    else:
                        if status == 1:
                            self.m_image_out_port.add_attribute(key, values, static=False)

                        else:
                            for j in values:
                                self.m_image_out_port.append_attribute_data(key, j)

            count += nimages

        history = f'number of input tags = {np.size(self.m_image_in_tags)}'
        self.m_image_out_port.add_history('CombineTagsModule', history)
        self.m_image_out_port.close_port()



import os

import pytest
import numpy as np

from pynpoint.core.dataio import DataStorage, InputPort, OutputPort
from pynpoint.util.tests import create_random, create_config, remove_test_data


class TestInputPort:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir)
        create_config(self.test_dir+'PynPoint_config.ini')

        file_in = os.path.dirname(__file__) + '/PynPoint_database.hdf5'

        self.storage = DataStorage(file_in)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_create_instance_access_data(self) -> None:

        with pytest.raises(ValueError) as error:
            InputPort('config', self.storage)

        assert str(error.value) == 'The tag name \'config\' is reserved for the central ' \
                                   'configuration of PynPoint.'

        with pytest.raises(ValueError) as error:
            InputPort('fits_header', self.storage)

        assert str(error.value) == 'The tag name \'fits_header\' is reserved for storage of the ' \
                                   'FITS headers.'

        port = InputPort('images', self.storage)

        assert port[0, 0, 0] == pytest.approx(0.00032486907273264834, rel=self.limit, abs=0.)

        data = np.mean(port.get_all())
        assert data == pytest.approx(1.1824138000882435e-05, rel=self.limit, abs=0.)

        assert len(port[0:2, 0, 0]) == 2
        assert port.get_shape() == (5, 11, 11)

        assert port.get_attribute('PIXSCALE') == 0.01
        assert port.get_attribute('PARANG')[0] == 0.

        with pytest.warns(UserWarning):
            assert port.get_attribute('none') is None

    def test_create_instance_access_non_existing_data(self) -> None:

        port = InputPort('test', self.storage)

        with pytest.warns(UserWarning):
            assert port[0, 0, 0] is None

        with pytest.warns(UserWarning):
            assert port.get_all() is None

        with pytest.warns(UserWarning):
            assert port.get_shape() is None

        with pytest.warns(UserWarning):
            assert port.get_attribute('num_files') is None

        with pytest.warns(UserWarning):
            assert port.get_all_non_static_attributes() is None

        with pytest.warns(UserWarning):
            assert port.get_all_static_attributes() is None

    def test_create_instance_no_data_storage(self) -> None:

        port = InputPort('test')

        with pytest.warns(UserWarning):
            assert port[0, 0, 0] is None

        with pytest.warns(UserWarning):
            assert port.get_all() is None

        with pytest.warns(UserWarning):
            assert port.get_shape() is None

        with pytest.warns(UserWarning):
            assert port.get_all_non_static_attributes() is None

        with pytest.warns(UserWarning):
            assert port.get_all_static_attributes() is None

    # def test_get_all_attributes(self) -> None:
    #
    #     port = InputPort('images', self.storage)
    #
    #     assert port.get_all_static_attributes() == {'PIXSCALE': 0.01}
    #     assert port.get_all_non_static_attributes() == ['PARANG', ]
    #
    #     port = OutputPort('images', self.storage)
    #     assert port.del_all_attributes() is None
    #
    #     port = InputPort('images', self.storage)
    #     assert port.get_all_non_static_attributes() is None

    # def test_get_ndim(self) -> None:
    #
    #     with pytest.warns(UserWarning) as warning:
    #         ndim = InputPort('images', None).get_ndim()
    #
    #     assert len(warning) == 1
    #
    #     assert warning[0].message.args[0] == 'InputPort can not load data ' \
    #                                          'unless a database is connected.'
    #
    #     assert ndim is None
    #
    #     port = InputPort('images', self.storage)
    #     assert port.get_ndim() == 3



"""
Pipeline modules for basic image operations.
"""

import time

from typing import Tuple

from typeguard import typechecked
from scipy.ndimage import rotate

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import progress, memory_frames


class SubtractImagesModule(ProcessingModule):
    """
    Pipeline module for subtracting two sets of images.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tags: Tuple[str, str],
                 image_out_tag: str,
                 scaling: float = 1.) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tags : tuple(str, str)
            Tuple with two tags of the database entry that are read as input.
        image_out_tag : str
            Tag of the database entry with the subtracted images that are written as output.
        scaling : float
            Additional scaling factor.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in1_port = self.add_input_port(image_in_tags[0])
        self.m_image_in2_port = self.add_input_port(image_in_tags[1])
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_scaling = scaling

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Subtracts the images from the second database tag from the images
        of the first database tag, on a frame-by-frame basis.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_in1_port.get_shape() != self.m_image_in2_port.get_shape():
            raise ValueError('The shape of the two input tags has to be the same.')

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in1_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Subtracting images...', start_time)

            images1 = self.m_image_in1_port[frames[i]:frames[i+1], ]
            images2 = self.m_image_in2_port[frames[i]:frames[i+1], ]

            self.m_image_out_port.append(self.m_scaling*(images1-images2), data_dim=3)

        history = f'scaling = {self.m_scaling}'
        self.m_image_out_port.add_history('SubtractImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in1_port)
        self.m_image_out_port.close_port()


class AddImagesModule(ProcessingModule):
    """
    Pipeline module for adding two sets of images.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tags: Tuple[str, str],
                 image_out_tag: str,
                 scaling: float = 1.) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tags : tuple(str, str)
            Tuple with two tags of the database entry that are read as input.
        image_out_tag : str
            Tag of the database entry with the added images that are written as output.
        scaling: float
            Additional scaling factor.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in1_port = self.add_input_port(image_in_tags[0])
        self.m_image_in2_port = self.add_input_port(image_in_tags[1])
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_scaling = scaling

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Add the images from the two database tags on a frame-by-frame
        basis.

        Returns
        -------
        NoneType
            None
        """

        if self.m_image_in1_port.get_shape() != self.m_image_in2_port.get_shape():
            raise ValueError('The shape of the two input tags has to be the same.')

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in1_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Adding images...', start_time)

            images1 = self.m_image_in1_port[frames[i]:frames[i+1], ]
            images2 = self.m_image_in2_port[frames[i]:frames[i+1], ]

            self.m_image_out_port.append(self.m_scaling*(images1+images2), data_dim=3)

        history = f'scaling = {self.m_scaling}'
        self.m_image_out_port.add_history('AddImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in1_port)
        self.m_image_out_port.close_port()


class RotateImagesModule(ProcessingModule):
    """
    Pipeline module for rotating images.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 angle: float) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        angle : float
            Rotation angle (deg). Rotation is clockwise for positive values.
        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_angle = angle

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Rotates all images by a constant angle.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Rotating images...', start_time)

            images = self.m_image_in_port[frames[i]:frames[i+1], ]

            for j in range(frames[i+1]-frames[i]):
                im_tmp = images[j, ]

                # ndimage.rotate rotates in clockwise direction for positive angles
                im_tmp = rotate(im_tmp, self.m_angle, reshape=False)

                self.m_image_out_port.append(im_tmp, data_dim=3)

        history = f'angle [deg] = {self.m_angle}'
        self.m_image_out_port.add_history('RotateImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()


class RepeatImagesModule(ProcessingModule):
    """
    Pipeline module for repeating the images from a dataset.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 repeat: int) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry with the added images that are written as output.
        repeat: int
            The number of times the input images get repeated.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_repeat = repeat

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Repeats the stack of input images a specified number of times.

        Returns
        -------
        NoneType
            None
        """

        nimages = self.m_image_in_port.get_shape()[0]
        memory = self._m_config_port.get_attribute('MEMORY')

        frames = memory_frames(memory, nimages)

        start_time = time.time()

        for i in range(self.m_repeat):
            progress(i, self.m_repeat, 'Repeating images...', start_time)

            for j, _ in enumerate(frames[:-1]):
                images = self.m_image_in_port[frames[j]:frames[j+1], ]
                self.m_image_out_port.append(images, data_dim=3)

        history = f'repeat = {self.m_repeat}'
        self.m_image_out_port.add_history('RepeatImagesModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()



"""
Pipeline modules for subtraction of the background emission.
"""

import time
import warnings

from typing import Any, Optional, Union

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.image import create_mask
from pynpoint.util.module import progress
from pynpoint.util.apply_func import subtract_line


class SimpleBackgroundSubtractionModule(ProcessingModule):
    """
    Pipeline module for simple background subtraction. Only applicable on data obtained with
    dithering.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 shift: int) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        shift : int
            Frame index offset for the background subtraction. Typically equal to the number of
            frames per dither location.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_shift = shift

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Simple background subtraction with a constant index offset.

        Returns
        -------
        NoneType
            None
        """

        nframes = self.m_image_in_port.get_shape()[0]

        subtract = self.m_image_in_port[0] - self.m_image_in_port[(0 + self.m_shift) % nframes]

        if self.m_image_in_port.tag == self.m_image_out_port.tag:
            self.m_image_out_port[0] = subtract
        else:
            self.m_image_out_port.set_all(subtract, data_dim=3)

        start_time = time.time()

        for i in range(1, nframes):
            progress(i, nframes, 'Subtracting background...', start_time)

            subtract = self.m_image_in_port[i] - self.m_image_in_port[(i + self.m_shift) % nframes]

            if self.m_image_in_port.tag == self.m_image_out_port.tag:
                self.m_image_out_port[i] = subtract
            else:
                self.m_image_out_port.append(subtract)

        history = f'shift = {self.m_shift}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('SimpleBackgroundSubtractionModule', history)
        self.m_image_out_port.close_port()


class MeanBackgroundSubtractionModule(ProcessingModule):
    """
    Pipeline module for mean background subtraction. Only applicable on data obtained with
    dithering.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 shift: Optional[int] = None,
                 cubes: int = 1) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output. Should be different from
            *image_in_tag*.
        shift : int, None
            Image index offset for the background subtraction. Typically equal to the number of
            frames per dither location. If set to None, the ``NFRAMES`` attribute will be used to
            select the background frames automatically. The *cubes* parameters should be set when
            *shift* is set to None.
        cubes : int
            Number of consecutive cubes per dithering position.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_shift = shift
        self.m_cubes = cubes

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Mean background subtraction which uses either a constant index
        offset or the ``NFRAMES`` attributes. The mean background is calculated from the cubes
        before and after the science cube.

        Returns
        -------
        NoneType
            None
        """

        # Use NFRAMES values if shift=None
        if self.m_shift is None:
            self.m_shift = self.m_image_in_port.get_attribute('NFRAMES')

        nframes = self.m_image_in_port.get_shape()[0]

        if not isinstance(self.m_shift, np.ndarray) and nframes < self.m_shift*2.0:
            raise ValueError('The input stack is too small for a mean background subtraction. The '
                             'position of the star should shift at least once.')

        if self.m_image_in_port.tag == self.m_image_out_port.tag:
            raise ValueError('The tag of the input port should be different from the output port.')

        # Number of substacks
        if isinstance(self.m_shift, np.ndarray):
            nstacks = np.size(self.m_shift)
        else:
            nstacks = int(np.floor(nframes/self.m_shift))

        # First mean subtraction to set up the output port array
        if isinstance(self.m_shift, np.ndarray):
            next_start = np.sum(self.m_shift[0:self.m_cubes])
            next_end = np.sum(self.m_shift[0:2*self.m_cubes])

            if 2*self.m_cubes > np.size(self.m_shift):
                raise ValueError('Not enough frames available for the background subtraction.')

            bg_data = self.m_image_in_port[next_start:next_end, ]
            bg_mean = np.mean(bg_data, axis=0)

        else:
            bg_data = self.m_image_in_port[self.m_shift:2*self.m_shift, ]
            bg_mean = np.mean(bg_data, axis=0)

        # Initiate the result port data with the first frame
        bg_sub = self.m_image_in_port[0, ] - bg_mean
        self.m_image_out_port.set_all(bg_sub, data_dim=3)

        # Mean subtraction of the first stack (minus the first frame)
        if isinstance(self.m_shift, np.ndarray):
            tmp_data = self.m_image_in_port[1:next_start, ]
            tmp_data = tmp_data - bg_mean
            self.m_image_out_port.append(tmp_data)

        else:
            tmp_data = self.m_image_in_port[1:self.m_shift, ]
            tmp_data = tmp_data - bg_mean
            self.m_image_out_port.append(tmp_data)

        # Processing of the rest of the data
        start_time = time.time()
        if isinstance(self.m_shift, np.ndarray):
            for i in range(self.m_cubes, nstacks, self.m_cubes):
                progress(i, nstacks, 'Subtracting background...', start_time)

                prev_start = np.sum(self.m_shift[0:i-self.m_cubes])
                prev_end = np.sum(self.m_shift[0:i])

                next_start = np.sum(self.m_shift[0:i+self.m_cubes])
                next_end = np.sum(self.m_shift[0:i+2*self.m_cubes])

                # calc the mean (previous)
                tmp_data = self.m_image_in_port[prev_start:prev_end, ]
                tmp_mean = np.mean(tmp_data, axis=0)

                if i < nstacks-self.m_cubes:
                    # calc the mean (next)
                    tmp_data = self.m_image_in_port[next_start:next_end, ]
                    tmp_mean = (tmp_mean + np.mean(tmp_data, axis=0)) / 2.0

                # subtract mean
                tmp_data = self.m_image_in_port[prev_end:next_start, ]
                tmp_data = tmp_data - tmp_mean
                self.m_image_out_port.append(tmp_data)

        else:
            # the last and the one before will be performed afterwards
            top = int(np.ceil(nframes/self.m_shift)) - 2

            for i in range(1, top, 1):
                progress(i, top, 'Subtracting background...', start_time)

                # calc the mean (next)
                tmp_data = self.m_image_in_port[(i+1)*self.m_shift:(i+2)*self.m_shift, ]
                tmp_mean = np.mean(tmp_data, axis=0)

                # calc the mean (previous)
                tmp_data = self.m_image_in_port[(i-1)*self.m_shift:(i+0)*self.m_shift, ]
                tmp_mean = (tmp_mean + np.mean(tmp_data, axis=0)) / 2.0

                # subtract mean
                tmp_data = self.m_image_in_port[(i+0)*self.m_shift:(i+1)*self.m_shift, ]
                tmp_data = tmp_data - tmp_mean
                self.m_image_out_port.append(tmp_data)

            # last and the one before
            # 1. ------------------------------- one before -------------------
            # calc the mean (previous)
            tmp_data = self.m_image_in_port[(top-1)*self.m_shift:(top+0)*self.m_shift, ]
            tmp_mean = np.mean(tmp_data, axis=0)

            # calc the mean (next)
            # 'nframes' is important if the last step is to huge
            tmp_data = self.m_image_in_port[(top+1)*self.m_shift:nframes, ]
            tmp_mean = (tmp_mean + np.mean(tmp_data, axis=0)) / 2.0

            # subtract mean
            tmp_data = self.m_image_in_port[top*self.m_shift:(top+1)*self.m_shift, ]
            tmp_data = tmp_data - tmp_mean
            self.m_image_out_port.append(tmp_data)

            # 2. ------------------------------- last -------------------
            # calc the mean (previous)
            tmp_data = self.m_image_in_port[(top+0)*self.m_shift:(top+1)*self.m_shift, ]
            tmp_mean = np.mean(tmp_data, axis=0)

            # subtract mean
            tmp_data = self.m_image_in_port[(top+1)*self.m_shift:nframes, ]
            tmp_data = tmp_data - tmp_mean
            self.m_image_out_port.append(tmp_data)
            # -----------------------------------------------------------

        if isinstance(self.m_shift, np.ndarray):
            history = 'shift = NFRAMES'
        else:
            history = f'shift = {self.m_shift}'

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('MeanBackgroundSubtractionModule', history)
        self.m_image_out_port.close_port()


class LineSubtractionModule(ProcessingModule):
    """
    Pipeline module for subtracting the background emission from each pixel by computing the mean
    or median of all values in the row and column of the pixel. The module can for example be
    used if no background data is available or to remove a detector bias.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 combine: str = 'median',
                 mask: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.
        combine : str
            The method by which the column and row pixel values are combined ('median' or 'mean').
            Using a mean-combination is computationally faster than a median-combination.
        mask : float, None
            The radius of the mask within which pixel values are ignored. No mask is used if set
            to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_combine = combine
        self.m_mask = mask

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Selects the pixel values in the column and row at each pixel
        position, computes the mean or median value while excluding pixels within the radius of
        the mask, and subtracts the mean or median value from each pixel separately.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        im_shape = self.m_image_in_port.get_shape()[-2:]

        if self.m_mask:
            size = (self.m_mask/pixscale, None)
        else:
            size = (None, None)

        mask = create_mask(im_shape, size)

        self.apply_function_to_images(subtract_line,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Background subtraction',
                                      func_args=(mask,
                                                 self.m_combine,
                                                 im_shape))

        history = f'combine = {self.m_combine}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('LineSubtractionModule', history)
        self.m_image_out_port.close_port()


class NoddingBackgroundModule(ProcessingModule):
    """
    Pipeline module for background subtraction of data obtained with nodding (e.g., NACO AGPM
    data). Before using this module, the sky images should be stacked with the StackCubesModule
    such that each image in the stack of sky images corresponds to the mean combination of a
    single FITS data cube.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 science_in_tag: str,
                 sky_in_tag: str,
                 image_out_tag: str,
                 mode: str = 'both') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        science_in_tag : str
            Tag of the database entry with science images that are read as input.
        sky_in_tag : str
            Tag of the database entry with sky images that are read as input. The
            :class:`~pynpoint.processing.stacksubset.StackCubesModule` should be used on the sky
            images beforehand.
        image_out_tag : str
            Tag of the database entry with sky subtracted images that are written as output.
        mode : str
            Sky images that are subtracted, relative to the science images. Either the next,
            previous, or average of the next and previous cubes of sky frames can be used by
            choosing 'next', 'previous', or 'both', respectively.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_science_in_port = self.add_input_port(science_in_tag)
        self.m_sky_in_port = self.add_input_port(sky_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_time_stamps = []

        if mode in ['next', 'previous', 'both']:
            self.m_mode = mode
        else:
            raise ValueError('Mode needs to be \'next\', \'previous\', or \'both\'.')

    @typechecked
    def _create_time_stamp_list(self) -> None:
        """
        Internal method for assigning a time stamp, based on the exposure number ID, to each cube
        of sky and science images.
        """

        class TimeStamp:
            """
            Class for creating a time stamp.
            """

            @typechecked
            def __init__(self,
                         time_in: Any,
                         im_type: str,
                         index: Union[int, slice]) -> None:

                self.m_time = time_in
                self.m_im_type = im_type
                self.m_index = index

            @typechecked
            def __repr__(self) -> str:

                return repr((self.m_time,
                             self.m_im_type,
                             self.m_index))

        exp_no_sky = self.m_sky_in_port.get_attribute('EXP_NO')
        exp_no_science = self.m_science_in_port.get_attribute('EXP_NO')

        nframes_sky = self.m_sky_in_port.get_attribute('NFRAMES')
        nframes_science = self.m_science_in_port.get_attribute('NFRAMES')

        if np.all(nframes_sky != 1):
            warnings.warn('The NFRAMES values of the sky images are not all equal to unity. '
                          'The StackCubesModule should be applied on the sky images before the '
                          'NoddingBackgroundModule is used.')

        for i, item in enumerate(exp_no_sky):
            self.m_time_stamps.append(TimeStamp(item, 'SKY', i))

        current = 0
        for i, item in enumerate(exp_no_science):
            frames = slice(current, current+nframes_science[i])
            self.m_time_stamps.append(TimeStamp(item, 'SCIENCE', frames))
            current += nframes_science[i]

        self.m_time_stamps = sorted(self.m_time_stamps, key=lambda time_stamp: time_stamp.m_time)

    @typechecked
    def calc_sky_frame(self,
                       index_of_science_data: int) -> np.ndarray:
        """
        Method for finding the required sky frame (next, previous, or the mean of next and
        previous) by comparing the time stamp of the science frame with preceding and following
        sky frames.
        """

        if not any(x.m_im_type == 'SKY' for x in self.m_time_stamps):
            raise ValueError('List of time stamps does not contain any SKY images.')

        @typechecked
        def search_for_next_sky() -> np.ndarray:
            for i in range(index_of_science_data, len(self.m_time_stamps)):
                if self.m_time_stamps[i].m_im_type == 'SKY':
                    return self.m_sky_in_port[self.m_time_stamps[i].m_index, ]

            # no next sky found, look for previous sky
            return search_for_previous_sky()

        @typechecked
        def search_for_previous_sky() -> np.ndarray:
            for i in reversed(list(range(0, index_of_science_data))):
                if self.m_time_stamps[i].m_im_type == 'SKY':
                    return self.m_sky_in_port[self.m_time_stamps[i].m_index, ]

            # no previous sky found, look for next sky
            return search_for_next_sky()

        if self.m_mode == 'next':
            return search_for_next_sky()

        if self.m_mode == 'previous':
            return search_for_previous_sky()

        if self.m_mode == 'both':
            previous_sky = search_for_previous_sky()
            next_sky = search_for_next_sky()

            return (previous_sky+next_sky)/2.

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Create list of time stamps, get sky and science images, and
        subtract the sky images from the science images.

        Returns
        -------
        NoneType
            None
        """

        self._create_time_stamp_list()

        start_time = time.time()
        for i, time_entry in enumerate(self.m_time_stamps):
            progress(i, len(self.m_time_stamps), 'Subtracting background...', start_time)

            if time_entry.m_im_type == 'SKY':
                continue

            sky = self.calc_sky_frame(i)
            science = self.m_science_in_port[time_entry.m_index, ]

            self.m_image_out_port.append(science - sky[None, ], data_dim=3)

        history = f'mode = {self.m_mode}'
        self.m_image_out_port.copy_attributes(self.m_science_in_port)
        self.m_image_out_port.add_history('NoddingBackgroundModule', history)
        self.m_image_out_port.close_port()



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.attr_writing import ParangWritingModule, AttributeWritingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestAttributeWriting:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir)
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=['parang.dat', 'attribute.dat'])

    def test_input_data(self) -> None:

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.007153603490533874, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_parang_writing(self) -> None:

        module = ParangWritingModule(file_name='parang.dat',
                                     name_in='parang_write1',
                                     output_dir=None,
                                     data_tag='images',
                                     header=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('parang_write1')

        data = np.loadtxt(self.test_dir+'parang.dat')
        assert np.sum(data) == pytest.approx(10., rel=self.limit, abs=0.)
        assert data.shape == (5, )

    def test_attribute_writing(self) -> None:

        module = AttributeWritingModule(file_name='attribute.dat',
                                        name_in='attr_write1',
                                        output_dir=None,
                                        data_tag='images',
                                        attribute='PARANG',
                                        header=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('attr_write1')

        data = np.loadtxt(self.test_dir+'attribute.dat')

        assert np.sum(data) == pytest.approx(10., rel=self.limit, abs=0.)
        assert data.shape == (5, )

    def test_attribute_not_present(self) -> None:

        module = AttributeWritingModule(file_name='attribute.dat',
                                        name_in='attr_write3',
                                        output_dir=None,
                                        data_tag='images',
                                        attribute='test',
                                        header=None)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('attr_write3')

        assert str(error.value) == 'The \'test\' attribute is not present in \'images\'.'

    def test_parang_writing_not_present(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            del hdf_file['header_images/PARANG']

        module = ParangWritingModule(file_name='parang.dat',
                                     name_in='parang_write3',
                                     output_dir=None,
                                     data_tag='images',
                                     header=None)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('parang_write3')

        assert str(error.value) == 'The PARANG attribute is not present in \'images\'.'



# This code is written by Davide Albanese, <albanese@fbk.eu>
# (C) 2011 mlpy Developers.

# See: Practical Guide to Wavelet Analysis - C. Torrence and G. P. Compo.

# Changes made by the PynPoint developers:
# - added type hints and type checks

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

import numpy as np

from scipy.special import gamma
from typeguard import typechecked


@typechecked
def normalization(s: Union[np.ndarray, np.generic],
                  dt: int) -> Union[np.ndarray, np.generic]:
    """"
    Parameters
    ----------
    s : numpy.ndarray
        Scales.
    dt : int
        Time step.

    Returns
    -------
    numpy.ndarray
        Normalized data.
    """

    return np.sqrt((2 * np.pi * s) / dt)


@typechecked
def morletft(s: np.ndarray,
             w: np.ndarray,
             w0: int,
             dt: int) -> np.ndarray:
    """"
    Fourier transformed morlet function.

    Parameters
    ----------
    s : numpy.ndarray
        Scales.
    w : numpy.ndarray
        Angular frequencies.
    w0 : int
        Omega0 frequency.
    dt : int
        Time step.

    Returns
    -------
    numpy.ndarray
        Normalized Fourier transformed morlet function
    """

    p = 0.75112554446494251  # pi**(-1.0/4.0)
    pos = w > 0
    wavelet = np.zeros((s.shape[0], w.shape[0]))

    for i in range(s.shape[0]):
        n = normalization(s[i], dt)
        wavelet[i][pos] = n * p * np.exp(-(s[i] * w[pos] - w0) ** 2 / 2.0)

    return wavelet


@typechecked
def dogft(s: np.ndarray,
          w: np.ndarray,
          order: int,
          dt: int) -> np.ndarray:
    """
    Fourier transformed DOG function.

    Parameters
    ----------
    s : numpy.ndarray
        Scales.
    w : numpy.ndarray
        Angular frequencies.
    order : int
        Wavelet order.
    dt : int
        Time step.

    Returns
    -------
    numpy.ndarray
        Normalized Fourier transformed DOG function.
    """

    p = - (0.0 + 1.0j) ** order / np.sqrt(gamma(order + 0.5))
    wavelet = np.zeros((s.shape[0], w.shape[0]), dtype=np.complex128)

    for i in range(s.shape[0]):
        n = normalization(s[i], dt)
        h = s[i] * w
        wavelet[i] = n * p * h ** order * np.exp(-h ** 2 / 2.0)

    return wavelet


@typechecked
def angularfreq(N: int,
                dt: int) -> np.ndarray:
    """
    Compute angular frequencies.

    Parameters
    ----------
    N : int
        Number of data samples.
    dt : int
        Time step.

    Returns
    -------
    numpy.ndarray
        Angular frequencies (1D).
    """

    # See (5) at page 64.

    N2 = int(N / 2.0)
    w = np.empty(N)

    for i in range(w.shape[0]):
        if i <= N2:
            w[i] = (2 * np.pi * i) / (N * dt)
        else:
            w[i] = (2 * np.pi * (i - N)) / (N * dt)

    return w


@typechecked
def autoscales(N: int,
               dt: int,
               dj: float,
               wf: str,
               p: int) -> np.ndarray:
    """
    Compute scales as fractional power of two.

    Parameters
    ----------
    N : int
        Number of data samples.
    dt : int
        Time step.
    dj : float
        Scale resolution (smaller values of give finer resolution).
    wf : str
        Wavelet function ("morlet", "paul", or "dog").
    p : int
        omega0 ("morlet") or order ("paul", "dog").

    Returns
    -------
    numpy.ndarray
        Scales (1D).
    """

    if wf == 'dog':
        s0 = (dt * np.sqrt(p + 0.5)) / np.pi

    elif wf == 'morlet':
        s0 = (dt * (p + np.sqrt(2 + p ** 2))) / (2 * np.pi)

    else:
        raise ValueError('Wavelet function not available.')

    # See (9) and (10) at page 67.

    J = int(np.floor(dj ** -1 * np.log2((N * dt) / s0)))
    s = np.empty(J + 1)

    for i in range(s.shape[0]):
        s[i] = s0 * 2 ** (i * dj)

    return s


# def fourier_from_scales(scales, wf, p):
#     """Compute the equivalent fourier period
#     from scales.
#
#     :Parameters:
#        scales : list or 1d numpy array
#           scales
#        wf : string ('morlet', 'paul', 'dog')
#           wavelet function
#        p : float
#           wavelet function parameter ('omega0' for morlet, 'm' for paul
#           and dog)
#
#     :Returns:
#        fourier wavelengths
#     """
#
#     scales_arr = np.asarray(scales)
#
#     if wf == 'dog':
#         return (2 * np.pi * scales_arr) / np.sqrt(p + 0.5)
#     elif wf == 'morlet':
#         return (4 * np.pi * scales_arr) / (p + np.sqrt(2 + p ** 2))
#     else:
#         raise ValueError('wavelet function not available')


# def scales_from_fourier(f, wf, p):
#     """Compute scales from fourier period.
#
#     :Parameters:
#        f : list or 1d numpy array
#           fourier wavelengths
#        wf : string ('morlet', 'paul', 'dog')
#           wavelet function
#        p : float
#           wavelet function parameter ('omega0' for morlet, 'm' for paul
#           and dog)
#
#     :Returns:
#        scales
#     """
#
#     f_arr = np.asarray(f)
#
#     if wf == 'dog':
#         return (f_arr * np.sqrt(p + 0.5)) / (2 * np.pi)
#     elif wf == 'morlet':
#         return (f_arr * (p + np.sqrt(2 + p ** 2))) / (4 * np.pi)
#     else:
#         raise ValueError('wavelet function not available')


@typechecked
def cwt(x: np.ndarray,
        dt: int,
        scales: np.ndarray,
        wf: str = "dog",
        p: int = 2) -> np.ndarray:
    """
    Continuous Wavelet Transform.

    Parameters
    ----------
    x : numpy.ndarray
        Data (1D).
    dt : int
        Time step.
    scales : numpy.ndarray
        Scales (1D).
    wf : str
        Wavelet function ("morlet", "paul", or "dog").
    p : int
        omega0 ("morlet") or order ("paul", "dog").

    Returns
    -------
    numpy.ndarray
        Transformed data (2D).
    """

    x_arr = np.asarray(x) - np.mean(x)
    scales_arr = np.asarray(scales)

    if x_arr.ndim != 1:
        raise ValueError('x must be an 1d numpy array of list')

    if scales_arr.ndim != 1:
        raise ValueError('scales must be an 1d numpy array of list')

    w = angularfreq(N=x_arr.shape[0], dt=dt)

    if wf == 'dog':
        wft = dogft(s=scales_arr, w=w, order=p, dt=dt)

    elif wf == 'morlet':
        wft = morletft(s=scales_arr, w=w, w0=p, dt=dt)

    else:
        raise ValueError('wavelet function is not available')

    X_ARR = np.empty((wft.shape[0], wft.shape[1]), dtype=np.complex128)

    x_arr_ft = np.fft.fft(x_arr)

    for i in range(X_ARR.shape[0]):
        X_ARR[i] = np.fft.ifft(x_arr_ft * wft[i])

    return X_ARR


@typechecked
def icwt(X: np.ndarray,
         scales: np.ndarray) -> np.ndarray:
    """
    Inverse Continuous Wavelet Transform. The reconstruction factor is not applied.

    Parameters
    ----------
    X : numpy.ndarray
        Transformed data (2D).
    scales : numpy.ndarray
        Scales (1D).

    Returns
    -------
    numpy.ndarray
         1D data.
    """

    X_arr = np.asarray(X)
    scales_arr = np.asarray(scales)

    if X_arr.shape[0] != scales_arr.shape[0]:
        raise ValueError('X, scales: shape mismatch')

    # See (11), (13) at page 68
    X_ARR = np.empty_like(X_arr)
    for i in range(scales_arr.shape[0]):
        X_ARR[i] = X_arr[i] / np.sqrt(scales_arr[i])

    return np.sum(np.real(X_ARR), axis=0)



"""
Continuous wavelet transform (CWT) and discrete wavelet transform (DWT) denoising for speckle
suppression in the time domain. The module can be used as additional preprocessing step. See
Bonse et al. (arXiv:1804.05063) more information.
"""

from typing import Union

import pywt

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import cwt_denoise_line_in_time, dwt_denoise_line_in_time, \
                                     normalization


class CwtWaveletConfiguration:
    """
    Configuration capsule for a CWT based time denoising. Standard configuration as in the
    original paper.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 wavelet: str = 'dog',
                 wavelet_order: int = 2,
                 keep_mean: bool = False,
                 resolution: float = 0.5) -> None:
        """
        Parameters
        ----------
        wavelet : str
            Wavelet.
        wavelet_order : int
            Wavelet order.
        keep_mean : bool
            Keep mean.
        resolution : float
            Resolution.

        Returns
        -------
        NoneType
            None
        """

        if wavelet not in ['dog', 'morlet']:
            raise ValueError('CWT supports only \'dog\' and \'morlet\' wavelets.')

        self.m_wavelet = wavelet
        self.m_wavelet_order = wavelet_order
        self.m_keep_mean = keep_mean
        self.m_resolution = resolution


class DwtWaveletConfiguration:
    """
    Configuration capsule for a DWT based time denoising. A cheap alternative of the CWT based
    wavelet denoising. However, the supported wavelets should perform worse compared to the
    CWT DOG wavelet.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 wavelet: str = 'db8') -> None:
        """
        Parameters
        ----------
        wavelet : str
            Wavelet.

        Returns
        -------
        NoneType
            None
        """

        # create list of supported wavelets
        supported = []
        for item in pywt.families():
            supported += pywt.wavelist(item)

        # check if wavelet is supported
        if wavelet not in supported:
            raise ValueError(f'DWT supports only {supported} as input wavelet.')

        self.m_wavelet = wavelet


class WaveletTimeDenoisingModule(ProcessingModule):
    """
    Pipeline module for speckle subtraction in the time domain by using CWT or DWT wavelet
    shrinkage. See Bonse et al. (arXiv:1804.05063) for details.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 wavelet_configuration: Union[CwtWaveletConfiguration, DwtWaveletConfiguration],
                 padding: str = 'zero',
                 median_filter: bool = False,
                 threshold_function: str = 'soft') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name for the pipeline module.
        image_in_tag : str
            Database tag with the input data.
        image_out_tag : str
            Database tag for the output data.
        wavelet_configuration : pynpoint.processing.timedenoising.CwtWaveletConfiguration or \
                                pynpoint.processing.timedenoising.DwtWaveletConfiguration
            Instance of :class:`~pynpoint.processing.timedenoising.DwtWaveletConfiguration` or
            :class:`~pynpoint.processing.timedenoising.CwtWaveletConfiguration` which contains the
            parameters for the wavelet transformation.
        padding : str
            Padding method (``'zero'``, ``'mirror'``, or ``'none'``).
        median_filter : bool
            Apply a median filter in time to remove outliers, for example due to cosmic rays.
        threshold_function : str
            Threshold function that is used for wavelet shrinkage in the wavelet space
            (``'soft'`` or ``'hard'``).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_wavelet_configuration = wavelet_configuration
        self.m_median_filter = median_filter

        assert padding in ['zero', 'mirror', 'none']
        self.m_padding = padding

        assert threshold_function in ['soft', 'hard']
        self.m_threshold_function = threshold_function == 'soft'

        assert isinstance(wavelet_configuration,
                          (DwtWaveletConfiguration, CwtWaveletConfiguration))

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Applies the time denoising for the lines in time in parallel.

        Returns
        -------
        NoneType
            None
        """

        if isinstance(self.m_wavelet_configuration, DwtWaveletConfiguration):

            if self.m_padding == 'const_mean':
                self.m_padding = 'constant'

            if self.m_padding == 'none':
                self.m_padding = 'periodic'

            self.apply_function_in_time(dwt_denoise_line_in_time,
                                        self.m_image_in_port,
                                        self.m_image_out_port,
                                        func_args=(self.m_threshold_function,
                                                   self.m_padding,
                                                   self.m_wavelet_configuration))

        elif isinstance(self.m_wavelet_configuration, CwtWaveletConfiguration):

            self.apply_function_in_time(cwt_denoise_line_in_time,
                                        self.m_image_in_port,
                                        self.m_image_out_port,
                                        func_args=(self.m_threshold_function,
                                                   self.m_padding,
                                                   self.m_median_filter,
                                                   self.m_wavelet_configuration))

        if self.m_threshold_function:
            history = 'threshold_function = soft'
        else:
            history = 'threshold_function = hard'

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('WaveletTimeDenoisingModule', history)
        self.m_image_out_port.close_port()


class TimeNormalizationModule(ProcessingModule):
    """
    Pipeline module for normalization of global brightness variations of the detector. See Bonse
    et al. (arXiv:1804.05063) for details.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name for the pipeline module.
        image_in_tag : str
            Database tag with the input data.
        image_out_tag : str
            Database tag for the output data.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module.

        Returns
        -------
        NoneType
            None
        """

        self.apply_function_to_images(normalization,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Time normalization')

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('TimeNormalizationModule', 'normalization = median')
        self.m_image_out_port.close_port()



"""
Modules for writing data as text file.
"""

import os

from typing import Optional

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import WritingModule


class TextWritingModule(WritingModule):
    """
    Module for writing a 1D or 2D data set from the central HDF5 database as text file.
    TextWritingModule is a :class:`pynpoint.core.processing.WritingModule` and supports
    the use of the Pypeline default output directory as well as a specified location.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str,
                 output_dir: Optional[str] = None,
                 header: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry from which data is exported.
        file_name : str
            Name of the output file.
        output_dir : str, None
            Output directory where the text file will be stored. If no path is specified then the
            Pypeline default output location is used.
        header : str, None
            Header that is written at the top of the text file.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, output_dir=output_dir)

        self.m_data_port = self.add_input_port(data_tag)

        self.m_file_name = file_name
        self.m_header = header

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Saves the specified data from the database to a text file.

        Returns
        -------
        NoneType
            None
        """

        if self.m_header is None:
            self.m_header = ''

        print('Writing text file...', end='')

        out_name = os.path.join(self.m_output_location, self.m_file_name)

        data = self.m_data_port.get_all()

        if data.ndim == 3 and data.shape[0] == 1:
            data = np.squeeze(data, axis=0)

        if data.ndim > 2:
            raise ValueError('Only 1D or 2D arrays can be written to a text file.')

        if data.dtype == 'int32' or data.dtype == 'int64':
            np.savetxt(out_name, data, header=self.m_header, comments='# ', fmt='%i')

        elif data.dtype == 'float32' or data.dtype == 'float64':
            np.savetxt(out_name, data, header=self.m_header, comments='# ')

        print(' [DONE]')

        self.m_data_port.close_port()



"""
Capsule for multiprocessing of the PSF subtraction with PCA. Residuals are created in parallel for
a range of principal components for which the PCA basis is required as input.
"""

import sys
import multiprocessing

from typing import List, Optional, Tuple, Union

import numpy as np

from typeguard import typechecked
from sklearn.decomposition import PCA

from pynpoint.core.dataio import OutputPort
from pynpoint.util.multiproc import TaskProcessor, TaskCreator, TaskWriter, TaskResult, \
                                    TaskInput, MultiprocessingCapsule, to_slice
from pynpoint.util.postproc import postprocessor
from pynpoint.util.residuals import combine_residuals


class PcaTaskCreator(TaskCreator):
    """
    The TaskCreator of the PCA multiprocessing. Creates one task for each principal component
    number. Does not require an input port since the data is directly given to the task processors.
    """

    @typechecked
    def __init__(self,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 num_proc: int,
                 pca_numbers: Union[np.ndarray, tuple]) -> None:
        """
        Parameters
        ----------
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Input task queue.
        num_proc : int
            Number of processors.
        pca_numbers : np.ndarray, tuple
            Principal components for which the residuals are computed.

        Returns
        -------
        NoneType
            None
        """

        super(PcaTaskCreator, self).__init__(None, tasks_queue_in, None, num_proc)

        self.m_pca_numbers = pca_numbers

    @typechecked
    def run(self) -> None:
        """
        Run method of PcaTaskCreator.

        Returns
        -------
        NoneType
            None
        """
        if isinstance(self.m_pca_numbers, tuple):
            for i, pca_first in enumerate(self.m_pca_numbers[0]):
                for j, pca_secon in enumerate(self.m_pca_numbers[1]):
                    parameters = (((i, i+1, None), (j, j+1, None), (None, None, None)), )
                    self.m_task_queue.put(TaskInput(tuple((pca_first, pca_secon)), parameters))

            self.create_poison_pills()

        else:
            for i, pca_number in enumerate(self.m_pca_numbers):
                parameters = (((i, i+1, None), (None, None, None), (None, None, None)), )
                self.m_task_queue.put(TaskInput(pca_number, parameters))

            self.create_poison_pills()


class PcaTaskProcessor(TaskProcessor):
    """
    The TaskProcessor of the PCA multiprocessing is the core of the parallelization. An instance
    of this class will calculate one forward and backward PCA transformation given the pre-trained
    scikit-learn PCA model. It does not get data from the TaskCreator but uses its own copy of the
    input data, which are the same and independent for each task. The following residuals can be
    created:

    * Mean residuals -- requirements[0] = True
    * Median residuals -- requirements[1] = True
    * Noise-weighted residuals -- requirements[2] = True
    * Clipped mean of the residuals -- requirements[3] = True
    """

    @typechecked
    def __init__(self,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 result_queue_in: multiprocessing.JoinableQueue,
                 star_reshape: np.ndarray,
                 angles: np.ndarray,
                 scales: Optional[np.ndarray],
                 pca_model: Optional[PCA],
                 im_shape: tuple,
                 indices: Optional[np.ndarray],
                 requirements: Tuple[bool, bool, bool, bool],
                 processing_type: str) -> None:
        """
        Parameters
        ----------
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Input task queue.
        result_queue_in : multiprocessing.queues.JoinableQueue
            Input result queue.
        star_reshape : np.ndarray
            Reshaped (2D) stack of images.
        angles : np.ndarray
            Derotation angles (deg).
        scales : np.ndarray
            scaling factors
        pca_model : sklearn.decomposition.pca.PCA
            PCA object with the basis.
        im_shape : tuple(int, int, int)
            Original shape of the stack of images.
        indices : np.ndarray
            Non-masked image indices.
        requirements : tuple(bool, bool, bool, bool)
            Required output residuals.
        processing_type : str
            selected processing type.

        Returns
        -------
        NoneType
            None
        """

        super(PcaTaskProcessor, self).__init__(tasks_queue_in, result_queue_in)

        self.m_star_reshape = star_reshape
        self.m_pca_model = pca_model
        self.m_angles = angles
        self.m_scales = scales
        self.m_im_shape = im_shape
        self.m_indices = indices
        self.m_requirements = requirements
        self.m_processing_type = processing_type

    @typechecked
    def run_job(self,
                tmp_task: TaskInput) -> TaskResult:
        """
        Run method of PcaTaskProcessor.

        Parameters
        ----------
        tmp_task : pynpoint.util.multiproc.TaskInput
            Input task.

        Returns
        -------
        pynpoint.util.multiproc.TaskResult
            Output residuals.
        """

        # correct data type of pca_number if necessary
        if isinstance(tmp_task.m_input_data, tuple):
            pca_number = tmp_task.m_input_data
        else:
            pca_number = int(tmp_task.m_input_data)

        residuals, res_rot = postprocessor(images=self.m_star_reshape,
                                           angles=self.m_angles,
                                           scales=self.m_scales,
                                           pca_number=pca_number,
                                           pca_sklearn=self.m_pca_model,
                                           im_shape=self.m_im_shape,
                                           indices=self.m_indices,
                                           processing_type=self.m_processing_type)

        # differentiate between IFS data or Mono-Wavelength data
        if res_rot.ndim == 3:
            res_output = np.zeros((4, res_rot.shape[-2], res_rot.shape[-1]))

        else:
            res_output = np.zeros((4, len(self.m_star_reshape),
                                   res_rot.shape[-2], res_rot.shape[-1]))

        if self.m_requirements[0]:
            res_output[0, ] = combine_residuals(method='mean',
                                                res_rot=res_rot)

        if self.m_requirements[1]:
            res_output[1, ] = combine_residuals(method='median',
                                                res_rot=res_rot)

        if self.m_requirements[2]:
            res_output[2, ] = combine_residuals(method='weighted',
                                                res_rot=res_rot,
                                                residuals=residuals,
                                                angles=self.m_angles)

        if self.m_requirements[3]:
            res_output[3, ] = combine_residuals(method='clipped',
                                                res_rot=res_rot)

        sys.stdout.write('.')
        sys.stdout.flush()

        return TaskResult(res_output, tmp_task.m_job_parameter[0])


class PcaTaskWriter(TaskWriter):
    """
    The TaskWriter of the PCA parallelization. Four different ports are used to save the
    results of the task processors (mean, median, weighted, and clipped).
    """

    @typechecked
    def __init__(self,
                 result_queue_in: multiprocessing.JoinableQueue,
                 mean_out_port: Optional[OutputPort],
                 median_out_port: Optional[OutputPort],
                 weighted_out_port: Optional[OutputPort],
                 clip_out_port: Optional[OutputPort],
                 data_mutex_in: multiprocessing.Lock,
                 requirements: Tuple[bool, bool, bool, bool]) -> None:
        """
        Constructor of PcaTaskWriter.

        Parameters
        ----------
        result_queue_in : multiprocessing.queues.JoinableQueue
            Input result queue.
        mean_out_port : pynpoint.core.dataio.OutputPort
            Output port with the mean residuals. Not used if set to None.
        median_out_port : pynpoint.core.dataio.OutputPort
            Output port with the median residuals. Not used if set to None.
        weighted_out_port : pynpoint.core.dataio.OutputPort
            Output port with the noise-weighted residuals. Not used if set to None.
        clip_out_port : pynpoint.core.dataio.OutputPort
            Output port with the clipped mean residuals. Not used if set to None.
        data_mutex_in : multiprocessing.synchronize.Lock
            A mutual exclusion variable which ensure that no read and write simultaneously occur.
        requirements : tuple(bool, bool, bool, bool)
            Required output residuals.

        Returns
        -------
        NoneType
            None
        """

        super(PcaTaskWriter, self).__init__(result_queue_in, None, data_mutex_in)

        self.m_mean_out_port = mean_out_port
        self.m_median_out_port = median_out_port
        self.m_weighted_out_port = weighted_out_port
        self.m_clip_out_port = clip_out_port
        self.m_requirements = requirements

    @typechecked
    def run(self) -> None:
        """
        Run method of PcaTaskWriter. Writes the residuals to the output ports.

        Returns
        -------
        NoneType
            None
        """

        while True:
            next_result = self.m_result_queue.get()
            poison_pill_case = self.check_poison_pill(next_result)

            if poison_pill_case == 1:
                break

            if poison_pill_case == 2:
                continue

            with self.m_data_mutex:
                res_slice = to_slice(next_result.m_position)
                if next_result.m_position[1][0] is None:
                    res_slice = (next_result.m_position[0][0])
                else:
                    res_slice = (next_result.m_position[0][0], next_result.m_position[1][0])

                if self.m_requirements[0]:
                    self.m_mean_out_port._check_status_and_activate()
                    self.m_mean_out_port[res_slice] = next_result.m_data_array[0]
                    self.m_mean_out_port.close_port()

                if self.m_requirements[1]:
                    self.m_median_out_port._check_status_and_activate()
                    self.m_median_out_port[res_slice] = next_result.m_data_array[1]
                    self.m_median_out_port.close_port()

                if self.m_requirements[2]:
                    self.m_weighted_out_port._check_status_and_activate()
                    self.m_weighted_out_port[res_slice] = next_result.m_data_array[2]
                    self.m_weighted_out_port.close_port()

                if self.m_requirements[3]:
                    self.m_clip_out_port._check_status_and_activate()
                    self.m_clip_out_port[res_slice] = next_result.m_data_array[3]
                    self.m_clip_out_port.close_port()

            self.m_result_queue.task_done()


class PcaMultiprocessingCapsule(MultiprocessingCapsule):
    """
    Capsule for PCA multiprocessing with the poison pill pattern.
    """

    @typechecked
    def __init__(self,
                 mean_out_port: Optional[OutputPort],
                 median_out_port: Optional[OutputPort],
                 weighted_out_port: Optional[OutputPort],
                 clip_out_port: Optional[OutputPort],
                 num_proc: int,
                 pca_numbers: Union[tuple, np.ndarray],
                 pca_model: Optional[PCA],
                 star_reshape: np.ndarray,
                 angles: np.ndarray,
                 scales: Optional[np.ndarray],
                 im_shape: tuple,
                 indices: Optional[np.ndarray],
                 processing_type: str) -> None:
        """
        Constructor of PcaMultiprocessingCapsule.

        Parameters
        ----------
        mean_out_port : pynpoint.core.dataio.OutputPort
            Output port for the mean residuals.
        median_out_port : pynpoint.core.dataio.OutputPort
            Output port for the median residuals.
        weighted_out_port : pynpoint.core.dataio.OutputPort
            Output port for the noise-weighted residuals.
        clip_out_port : pynpoint.core.dataio.OutputPort
            Output port for the mean clipped residuals.
        num_proc : int
            Number of processors.
        pca_numbers : np.ndarray
            Number of principal components.
        pca_model : sklearn.decomposition.pca.PCA
            PCA object with the basis.
        star_reshape : np.ndarray
            Reshaped (2D) input images.
        angles : np.ndarray
            Derotation angles (deg).
        scales : np.ndarray
            scaling factors.
        im_shape : tuple(int, int, int)
            Original shape of the input images.
        indices : np.ndarray
            Non-masked pixel indices.
        processing_type : str
            selection of processing type

        Returns
        -------
        NoneType
            None
        """

        self.m_mean_out_port = mean_out_port
        self.m_median_out_port = median_out_port
        self.m_weighted_out_port = weighted_out_port
        self.m_clip_out_port = clip_out_port
        self.m_pca_numbers = pca_numbers
        self.m_pca_model = pca_model
        self.m_star_reshape = star_reshape
        self.m_angles = angles
        self.m_scales = scales
        self.m_im_shape = im_shape
        self.m_indices = indices
        self.m_processing_type = processing_type

        self.m_requirements = [False, False, False, False]

        if self.m_mean_out_port is not None:
            self.m_requirements[0] = True

        if self.m_median_out_port is not None:
            self.m_requirements[1] = True

        if self.m_weighted_out_port is not None:
            self.m_requirements[2] = True

        if self.m_clip_out_port is not None:
            self.m_requirements[3] = True

        self.m_requirements = tuple(self.m_requirements)

        super(PcaMultiprocessingCapsule, self).__init__(None, None, num_proc)

    @typechecked
    def create_writer(self,
                      image_out_port: None) -> PcaTaskWriter:
        """
        Method to create an instance of PcaTaskWriter.

        Parameters
        ----------
        image_out_port : None
            Output port, not used.

        Returns
        -------
        pynpoint.util.multipca.PcaTaskWriter
            PCA task writer.
        """

        return PcaTaskWriter(self.m_result_queue,
                             self.m_mean_out_port,
                             self.m_median_out_port,
                             self.m_weighted_out_port,
                             self.m_clip_out_port,
                             self.m_data_mutex,
                             self.m_requirements)

    @typechecked
    def init_creator(self,
                     image_in_port: None) -> PcaTaskCreator:
        """
        Method to create an instance of PcaTaskCreator.

        Parameters
        ----------
        image_in_port : None
            Input port, not used.

        Returns
        -------
        pynpoint.util.multipca.PcaTaskCreator
            PCA task creator.
        """

        return PcaTaskCreator(self.m_tasks_queue,
                              self.m_num_proc,
                              self.m_pca_numbers)

    @typechecked
    def create_processors(self) -> List[PcaTaskProcessor]:
        """
        Method to create a list of instances of PcaTaskProcessor.

        Returns
        -------
        list(pynpoint.util.multipca.PcaTaskProcessor, )
            PCA task processors.
        """

        processors = []

        for _ in range(self.m_num_proc):

            processors.append(PcaTaskProcessor(self.m_tasks_queue,
                                               self.m_result_queue,
                                               self.m_star_reshape,
                                               self.m_angles,
                                               self.m_scales,
                                               self.m_pca_model,
                                               self.m_im_shape,
                                               self.m_indices,
                                               self.m_requirements,
                                               self.m_processing_type))

        return processors



"""
Utilities for multiprocessing of lines in time with the poison pill pattern.
"""

import multiprocessing

from typing import Callable, List, Optional

import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import InputPort, OutputPort
from pynpoint.util.multiproc import TaskInput, TaskResult, TaskCreator, TaskProcessor, \
                                    MultiprocessingCapsule, apply_function


class LineReader(TaskCreator):
    """
    Reader of task inputs for :class:`~pynpoint.util.multiline.LineProcessingCapsule`. Continuously
    read all rows of a dataset and puts them into a task queue.
    """

    @typechecked
    def __init__(self,
                 data_port_in: InputPort,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 data_mutex_in: multiprocessing.Lock,
                 num_proc: int,
                 data_length: int) -> None:
        """
        Parameters
        ----------
        data_port_in : pynpoint.core.dataio.InputPort
            Input port.
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Tasks queue.
        data_mutex_in : multiprocessing.synchronize.Lock
            A mutex shared with the writer to ensure that no read and write operations happen at
            the same time.
        num_proc : int
            Number of processors.
        data_length : int
            Length of the processed data.

        Returns
        -------
        NoneType
            None
        """

        super(LineReader, self).__init__(data_port_in, tasks_queue_in, data_mutex_in, num_proc)

        self.m_data_length = data_length

    @typechecked
    def run(self) -> None:
        """
        Returns
        -------
        NoneType
            None
        """

        n_rows = self.m_data_in_port.get_shape()[1]
        row_length = int(np.ceil(self.m_data_in_port.get_shape()[1]/float(self.m_num_proc)))

        i = 0
        while i < n_rows:
            j = min((i + row_length), n_rows)

            # lock mutex and read data
            with self.m_data_mutex:
                self.m_data_in_port._check_status_and_activate()
                tmp_data = self.m_data_in_port[:, i:j, :]  # read rows from i to j
                self.m_data_in_port.close_port()

            param = (self.m_data_length, ((None, None, None), (i, j, None), (None, None, None)))
            self.m_task_queue.put(TaskInput(tmp_data, param))

            i = j

        self.create_poison_pills()


class LineTaskProcessor(TaskProcessor):
    """
    Processor of task inputs for :class:`~pynpoint.util.multiline.LineProcessingCapsule`. A
    processor applies a function on a row of lines in time.
    """

    @typechecked
    def __init__(self,
                 tasks_queue_in: multiprocessing.JoinableQueue,
                 result_queue_in: multiprocessing.JoinableQueue,
                 function: Callable,
                 function_args: Optional[tuple]) -> None:
        """
        Parameters
        ----------
        tasks_queue_in : multiprocessing.queues.JoinableQueue
            Tasks queue.
        result_queue_in : multiprocessing.queues.JoinableQueue
            Results queue.
        function : function
            Input function.
        function_args : tuple, None
            Optional function arguments.

        Returns
        -------
        NoneType
            None
        """

        super(LineTaskProcessor, self).__init__(tasks_queue_in, result_queue_in)

        self.m_function = function
        self.m_function_args = function_args

    @typechecked
    def run_job(self,
                tmp_task: TaskInput) -> TaskResult:
        """
        Parameters
        ----------
        tmp_task : pynpoint.util.multiproc.TaskInput
            Task input with the subsets of lines and the job parameters.

        Returns
        -------
        pynpoint.util.multiproc.TaskResult
            Task result.
        """

        result_arr = np.zeros((tmp_task.m_job_parameter[0],
                               tmp_task.m_input_data.shape[1],
                               tmp_task.m_input_data.shape[2]))

        count = 0

        for i in range(tmp_task.m_input_data.shape[1]):
            for j in range(tmp_task.m_input_data.shape[2]):
                result_arr[:, i, j] = apply_function(tmp_task.m_input_data[:, i, j], count,
                                                     self.m_function, self.m_function_args)

                count += 1

        return TaskResult(result_arr, tmp_task.m_job_parameter[1])


class LineProcessingCapsule(MultiprocessingCapsule):
    """
    Capsule for parallel processing of lines in time with the poison pill pattern. A function is
    applied in parallel to each line in time, for example as in
    :class:`~pynpoint.processing.timedenoising.WaveletTimeDenoisingModule`.
    """

    @typechecked
    def __init__(self,
                 image_in_port: InputPort,
                 image_out_port: OutputPort,
                 num_proc: int,
                 function: Callable,
                 function_args: Optional[tuple],
                 data_length: int) -> None:
        """
        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort
            Input port.
        image_out_port : pynpoint.core.dataio.OutputPort
            Output port.
        num_proc : int
            Number of processors.
        function : function
            Input function that is applied to the lines.
        function_args : tuple, None, optional
            Function arguments.
        data_length : int
            Length of the processed data.

        Returns
        -------
        NoneType
            None
        """

        self.m_function = function
        self.m_function_args = function_args
        self.m_data_length = data_length

        super(LineProcessingCapsule, self).__init__(image_in_port, image_out_port, num_proc)

    @typechecked
    def create_processors(self) -> List[LineTaskProcessor]:
        """
        Returns
        -------
        list(pynpoint.util.multiproc.LineTaskProcessor, )
            List with instances of :class:`~pynpoint.util.multiproc.LineTaskProcessor`
        """

        processors = []

        for _ in range(self.m_num_proc):

            processors.append(LineTaskProcessor(tasks_queue_in=self.m_tasks_queue,
                                                result_queue_in=self.m_result_queue,
                                                function=self.m_function,
                                                function_args=self.m_function_args))

        return processors

    @typechecked
    def init_creator(self,
                     image_in_port: InputPort) -> LineReader:
        """
        Parameters
        ----------
        image_in_port : pynpoint.core.dataio.InputPort
            Input port from where the subsets of lines are read.

        Returns
        -------
        pynpoint.util.multiline.LineReader
            Line reader object.
        """

        return LineReader(image_in_port,
                          self.m_tasks_queue,
                          self.m_data_mutex,
                          self.m_num_proc,
                          self.m_data_length)



"""
Pipeline modules for locating and extracting the position of a star.
"""

import math
import warnings

from typing import Optional, Tuple, Union

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import crop_around_star, crop_rotating_star
from pynpoint.util.image import rotate_coordinates


class StarExtractionModule(ProcessingModule):
    """
    Pipeline module to locate the position of the star in each image and to crop all the images
    around this position.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 index_out_tag: Optional[str] = None,
                 image_size: float = 2.,
                 fwhm_star: float = 0.2,
                 position: Optional[Union[Tuple[int, int, float],
                                          Tuple[None, None, float]]] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the dataset with the input images.
        image_out_tag : str
            Tag of the dataset that is stored as output, containing the extracted images.
        index_out_tag : str, None
            List with image indices for which the image size is too large to be cropped around the
            brightest pixel. No data is written if set to None. This tag name can be provided to
            the ``frames``` parameter in
            :class:`~pynpoint.processing.frameselection.RemoveFramesModule`. This argument is
            ignored if ``CPU`` is set to a value larger than 1.
        image_size : float
            Cropped image size (arcsec).
        fwhm_star : float
            Full width at half maximum (arcsec) of the Gaussian kernel that is used to smooth the
            images to lower contributions of bad pixels.
        position : tuple(int, int, float), None
            Subframe that is selected to search for the star. The tuple should contain a position
            (pix) and size (arcsec) as (pos_x, pos_y, size). The full image is used if set to None.
            The center of the image will be used with ``position=(None, None, size)``.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        if index_out_tag is None:
            self.m_index_out_port = None
        else:
            self.m_index_out_port = self.add_output_port(index_out_tag)

        self.m_image_size = image_size
        self.m_fwhm_star = fwhm_star
        self.m_position = position

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Locates the position of the star (only pixel precision) by
        selecting the highest pixel value. A Gaussian kernel with a FWHM similar to the PSF is
        used to lower the contribution of bad pixels which may have higher values than the peak
        of the PSF. Images are cropped and written to an output port.

        Returns
        -------
        NoneType
            None
        """

        cpu = self._m_config_port.get_attribute('CPU')

        if cpu > 1 and self.m_index_out_port is not None:
            warnings.warn('The \'index_out_port\' can only be used if CPU = 1. No data will '
                          'be stored to this output port.')

            del self._m_output_ports[self.m_index_out_port.tag]
            self.m_index_out_port = None

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        self.m_image_size = int(math.ceil(self.m_image_size/pixscale))
        self.m_fwhm_star = int(math.ceil(self.m_fwhm_star/pixscale))

        self.apply_function_to_images(crop_around_star,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Extracting stellar position',
                                      func_args=(self.m_position,
                                                 self.m_image_size,
                                                 self.m_fwhm_star,
                                                 pixscale,
                                                 self.m_index_out_port,
                                                 self.m_image_out_port))

        history = f'fwhm_star (pix) = {self.m_fwhm_star}'

        if self.m_index_out_port is not None:
            self.m_index_out_port.copy_attributes(self.m_image_in_port)
            self.m_index_out_port.add_history('StarExtractionModule', history)

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('StarExtractionModule', history)

        self.m_image_out_port.close_port()


class ExtractBinaryModule(ProcessingModule):
    """
    Pipeline module to extract a binary star (or another point source) which is rotating across the
    image stack.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 pos_center: Tuple[float, float],
                 pos_binary: Tuple[float, float],
                 image_size: float = 2.,
                 search_size: float = 0.1,
                 filter_size: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the dataset with the input images.
        image_out_tag : str
            Tag of the dataset that is stored as output, containing the extracted images.
        pos_center : tuple(float, float)
            Approximate position (x, y) of the center of rotation (pix).
        pos_binary : tuple(float, float)
            Approximate position (x, y) of the binary star in the first image (pix).
        image_size : float
            Cropped image size (arcsec).
        search_size : float
            Window size (arcsec) in which the brightest pixel is selected as position of the binary
            star. The search window is centered on the position that for each image is calculated
            from the ``pos_center``, ``pos_binary``, and parallactic angle (``PARANG``) of the
            image.
        filter_size : float, None
            Full width at half maximum (arcsec) of the Gaussian kernel that is used to smooth the
            images to lower contributions of bad pixels.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_pos_center = (pos_center[1], pos_center[0])  # (y, x)
        self.m_pos_binary = (pos_binary[1], pos_binary[0])  # (y, x)

        self.m_image_size = image_size
        self.m_search_size = search_size
        self.m_filter_size = filter_size

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Locates the position of a binary star (or some other point
        source) which rotates across the stack of images due to parallactic rotation. The
        approximate position of the binary star is calculated by taking into account the
        parallactic angle of each image separately. The brightest pixel is then selected as
        center around which the image is cropped.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        parang = self.m_image_in_port.get_attribute('PARANG')

        positions = np.zeros((parang.shape[0], 2), dtype=int)

        for i, item in enumerate(parang):
            # rotates in counterclockwise direction, hence the minus sign in angle
            positions[i, :] = rotate_coordinates(center=self.m_pos_center,
                                                 position=self.m_pos_binary,
                                                 angle=item-parang[0])

        self.m_image_size = int(math.ceil(self.m_image_size/pixscale))
        self.m_search_size = int(math.ceil(self.m_search_size/pixscale))

        if self.m_filter_size is not None:
            self.m_filter_size = int(math.ceil(self.m_filter_size/pixscale))

        self.apply_function_to_images(crop_rotating_star,
                                      self.m_image_in_port,
                                      self.m_image_out_port,
                                      'Extracting binary position',
                                      func_args=(positions,
                                                 self.m_image_size,
                                                 self.m_filter_size,
                                                 self.m_search_size))

        history = f'filter (pix) = {self.m_filter_size}'
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('ExtractBinaryModule', history)
        self.m_image_out_port.close_port()



"""
Module for writing a list of tags from the database to a separate HDF5 file.
"""

import os

from typing import Optional

import h5py

from typeguard import typechecked

from pynpoint.core.processing import WritingModule


class Hdf5WritingModule(WritingModule):
    """
    Module which exports a part of the PynPoint internal database to a separate HDF5 file. The
    datasets of the database can be chosen using the *tag_dictionary*. The module will also export
    the static and non-static attributes.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 file_name: str,
                 output_dir: Optional[str] = None,
                 tag_dictionary: Optional[dict] = None,
                 keep_attributes: bool = True,
                 overwrite: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        file_name : str
            Name of the file which will be created by the module.
        output_dir : str, None
            Location where the HDF5 file will be stored. The Pypeline default output location is
            used when no location is given.
        tag_dictionary : dict, None
            Directory containing all tags / keys of the datasets which will be exported from the
            PynPoint internal database. The datasets will be exported as
            {*input_tag*:*output_tag*, }.
        keep_attributes : bool
            If True all static and non-static attributes will be exported.
        overwrite : bool
            Overwrite an existing HDF5 file.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, output_dir=output_dir)

        if tag_dictionary is None:
            tag_dictionary = {}

        self.m_file_name = file_name
        self.m_tag_dictionary = tag_dictionary
        self.m_keep_attributes = keep_attributes
        self.m_overwrite = overwrite

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Exports all datasets defined in the *tag_dictionary* to an
        external HDF5 file.

        Returns
        -------
        NoneType
            None
        """

        print('Writing HDF5 file...', end='')

        if self.m_overwrite:
            out_file = h5py.File(os.path.join(self.m_output_location, self.m_file_name), mode='w')
        else:
            out_file = h5py.File(os.path.join(self.m_output_location, self.m_file_name), mode='a')

        for in_tag, out_tag in self.m_tag_dictionary.items():
            tmp_port = self.add_input_port(in_tag)
            tmp_data = tmp_port.get_all()

            if tmp_data is None:
                continue

            data_set = out_file.create_dataset(out_tag, data=tmp_data)

            if self.m_keep_attributes:
                # static attributes
                tmp_attr = tmp_port.get_all_static_attributes()

                # it is not possible to copy attributes all together
                for key, value in tmp_attr.items():
                    data_set.attrs[key] = value

                # non-static attributes
                non_static_attr_keys = tmp_port.get_all_non_static_attributes()

                if non_static_attr_keys is not None:
                    for key in non_static_attr_keys:
                        tmp_data_attr = tmp_port.get_attribute(key)
                        attr_tag = 'header_' + out_tag + '/' + key
                        out_file.create_dataset(attr_tag, data=tmp_data_attr)

            tmp_port.close_port()

        out_file.close()

        print(' [DONE]')



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.background import LineSubtractionModule
from pynpoint.processing.badpixel import BadPixelSigmaFilterModule
from pynpoint.processing.extract import StarExtractionModule
from pynpoint.processing.timedenoising import TimeNormalizationModule
from pynpoint.util.tests import create_config, create_star_data, remove_test_data


class TestProcessing:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        np.random.seed(1)

        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        large_data = np.random.normal(loc=0, scale=2e-4, size=(10000, 5, 5))

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'w') as hdf_file:
            hdf_file.create_dataset('images', data=images)
            hdf_file.create_dataset('large_data', data=large_data)

        create_star_data(path=self.test_dir+'images')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        self.pipeline.set_attribute('images', 'PIXSCALE', 0.1, static=True)
        self.pipeline.set_attribute('large_data', 'PIXSCALE', 0.1, static=True)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['images'])

    def test_output_port_name(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='images',
                                   input_dir=self.test_dir+'images')

        module.add_output_port('test')

        with pytest.warns(UserWarning) as warning:
            module.add_output_port('test')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'Tag \'test\' of ReadingModule \'read\' is already ' \
                                             'used.'

        module = BadPixelSigmaFilterModule(name_in='badpixel',
                                           image_in_tag='images',
                                           image_out_tag='im_out')

        module.add_output_port('test')

        with pytest.warns(UserWarning) as warning:
            module.add_output_port('test')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'Tag \'test\' of ProcessingModule \'badpixel\' is ' \
                                             'already used.'

    def test_output_port_set_connection(self) -> None:

        self.pipeline.m_data_storage.open_connection()

        module = BadPixelSigmaFilterModule(name_in='badpixel2',
                                           image_in_tag='images',
                                           image_out_tag='im_out')

        self.pipeline.add_module(module)

        port = module.add_output_port('test1')

        self.pipeline.m_data_storage.close_connection()

    def test_apply_function(self) -> None:

        self.pipeline.set_attribute('config', 'MEMORY', 20, static=True)
        self.pipeline.set_attribute('config', 'CPU', 4, static=True)

        module = LineSubtractionModule(name_in='subtract',
                                       image_in_tag='images',
                                       image_out_tag='im_subtract',
                                       combine='mean',
                                       mask=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('subtract')

        data = self.pipeline.get_data('images')
        assert np.mean(data) == pytest.approx(1.1824138000882435e-05, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data('im_subtract')
        assert np.mean(data) == pytest.approx(-1.2544487946113274e-21, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_apply_function_args_none(self) -> None:

        module = TimeNormalizationModule(name_in='norm',
                                         image_in_tag='images',
                                         image_out_tag='im_norm')

        self.pipeline.add_module(module)
        self.pipeline.run_module('norm')

        data = self.pipeline.get_data('im_norm')
        assert np.mean(data) == pytest.approx(2.4012571778516812e-06, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_apply_function_args_none_memory_none(self) -> None:

        self.pipeline.set_attribute('config', 'MEMORY', 0, static=True)

        module = TimeNormalizationModule(name_in='norm_none',
                                         image_in_tag='images',
                                         image_out_tag='im_norm')

        self.pipeline.add_module(module)
        self.pipeline.run_module('norm_none')

        data = self.pipeline.get_data('im_norm')
        assert np.mean(data) == pytest.approx(2.4012571778516812e-06, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_apply_function_same_port(self) -> None:

        module = LineSubtractionModule(name_in='subtract_same',
                                       image_in_tag='im_subtract',
                                       image_out_tag='im_subtract',
                                       combine='mean',
                                       mask=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('subtract_same')

        data = self.pipeline.get_data('im_subtract')
        assert np.mean(data) == pytest.approx(-1.4336557652700885e-21, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_apply_function_args_none_memory_none_same_port(self) -> None:

        self.pipeline.set_attribute('config', 'MEMORY', 0, static=True)

        data = self.pipeline.get_data('images')
        assert np.mean(data) == pytest.approx(1.1824138000882435e-05, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        module = TimeNormalizationModule(name_in='norm_none_same',
                                         image_in_tag='images',
                                         image_out_tag='images')

        self.pipeline.add_module(module)
        self.pipeline.run_module('norm_none_same')

        data = self.pipeline.get_data('images')
        assert np.mean(data) == pytest.approx(2.4012571778516812e-06, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    # def test_apply_function_to_images_memory_none(self) -> None:
    #
    #     module = StarExtractionModule(name_in='extract',
    #                                   image_in_tag='im_subtract',
    #                                   image_out_tag='extract',
    #                                   index_out_tag=None,
    #                                   image_size=0.5,
    #                                   fwhm_star=0.1,
    #                                   position=(None, None, 0.1))
    #
    #     self.pipeline.add_module(module)
    #     self.pipeline.run_module('extract')
    #
    #     data = self.pipeline.get_data('extract')
    #     assert np.mean(data) == pytest.approx(1.8259937251367536e-05, rel=self.limit, abs=0.)
    #     assert data.shape == (5, 5, 5)

    # def test_multiproc_large_data(self) -> None:
    #
    #     self.pipeline.set_attribute('config', 'MEMORY', 1000, static=True)
    #     self.pipeline.set_attribute('config', 'CPU', 1, static=True)
    #
    #     module = LineSubtractionModule(name_in='subtract_single',
    #                                    image_in_tag='large_data',
    #                                    image_out_tag='im_sub_single',
    #                                    combine='mean',
    #                                    mask=None)
    #
    #     self.pipeline.add_module(module)
    #     self.pipeline.run_module('subtract_single')
    #
    #     self.pipeline.set_attribute('config', 'CPU', 4, static=True)
    #
    #     module = LineSubtractionModule(name_in='subtract_multi',
    #                                    image_in_tag='large_data',
    #                                    image_out_tag='im_sub_multi',
    #                                    combine='mean',
    #                                    mask=None)
    #
    #     self.pipeline.add_module(module)
    #     self.pipeline.run_module('subtract_multi')
    #
    #     data_single = self.pipeline.get_data('im_sub_single')
    #     data_multi = self.pipeline.get_data('im_sub_multi')
    #     assert data_single == pytest.approx(data_multi, rel=self.limit, abs=0.)
    #     assert data_single.shape == data_multi.shape



"""
Pipeline modules for PCA-based background subtraction.
"""

import math
import time
import warnings

from typing import List, Optional, Tuple, Union

import numpy as np

from scipy.optimize import curve_fit
from scipy.sparse.linalg import svds
from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.processing.psfpreparation import SortParangModule
from pynpoint.processing.resizing import CropImagesModule
from pynpoint.processing.stacksubset import CombineTagsModule
from pynpoint.util.module import memory_frames, progress
from pynpoint.util.star import locate_star


class PCABackgroundPreparationModule(ProcessingModule):
    """
    Pipeline module for preparing the images for a PCA-based background subtraction.
    """

    __author__ = 'Tomas Stolker, Silvan Hunziker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 star_out_tag: str,
                 subtracted_out_tag: str,
                 background_out_tag: str,
                 dither: Union[Tuple[int, int, int],
                               Tuple[int, None, Tuple[float, float]]],
                 combine: str = 'mean') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the pipeline module instance.
        image_in_tag : str
            Database tag with the images that are read as input.
        star_out_tag : str
            Database tag to store the images that contain the star.
        subtracted_out_tag : str
            Database tag to store the mean/median background subtracted images with the star.
        background_out_tag : str
            Database tag to store the images that contain only background emission.
        dither : tuple(int, int, int), tuple(int, None, tuple(float, float))
            Tuple with the parameters for separating the star and background frames. The tuple
            should contain three values, ``(positions, cubes, first)``, with ``positions`` the
            number of unique dithering position, ``cubes`` the number of consecutive cubes per
            dithering position, and ``first`` the index value of the first cube which contains the
            star (Python indexing starts at zero). Sorting is based on the ``DITHER_X`` and
            ``DITHER_Y`` attributes when ``cubes`` is set to None. In that case, the ``first``
            value should be a tuple with the ``DITHER_X`` and ``DITHER_Y`` values in which the star
            appears first.
        combine : str
            Method for combining the background images ('mean' or 'median').

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_star_out_port = self.add_output_port(star_out_tag)
        self.m_subtracted_out_port = self.add_output_port(subtracted_out_tag)
        self.m_background_out_port = self.add_output_port(background_out_tag)

        if len(dither) != 3:
            raise ValueError('The \'dither\' argument should contain three values.')

        self.m_dither = dither
        self.m_combine = combine

    @typechecked
    def _prepare(self) -> Tuple[np.ndarray, np.ndarray]:

        nframes = self.m_image_in_port.get_attribute('NFRAMES')

        cube_mean = np.zeros((nframes.shape[0],
                              self.m_image_in_port.get_shape()[2],
                              self.m_image_in_port.get_shape()[1]))

        count = 0
        for i, item in enumerate(nframes):
            if self.m_combine == 'mean':
                cube_mean[i, ] = np.mean(self.m_image_in_port[count:count+item, ], axis=0)

            elif self.m_combine == 'median':
                cube_mean[i, ] = np.median(self.m_image_in_port[count:count+item, ], axis=0)

            count += item

        if self.m_dither[1] is None:
            dither_x = self.m_image_in_port.get_attribute('DITHER_X')
            dither_y = self.m_image_in_port.get_attribute('DITHER_Y')

            star = np.logical_and(dither_x == self.m_dither[2][0],
                                  dither_y == self.m_dither[2][1])

            bg_frames = np.invert(star)

        else:
            bg_frames = np.ones(nframes.shape[0], dtype=bool)

            for i in range(self.m_dither[2]*self.m_dither[1],
                           np.size(nframes),
                           self.m_dither[1]*self.m_dither[0]):

                bg_frames[i:i+self.m_dither[1]] = False

        return bg_frames, cube_mean

    @typechecked
    def _separate(self,
                  bg_frames: np.ndarray,
                  bg_indices: np.ndarray,
                  parang: Optional[np.ndarray],
                  cube_mean: np.ndarray) -> Tuple[np.array, Optional[np.ndarray], np.ndarray,
                                                  np.ndarray, Optional[np.ndarray], np.ndarray]:

        @typechecked
        def _initialize() -> Tuple[np.array, Optional[np.ndarray], np.ndarray, np.ndarray,
                                   Optional[np.ndarray], np.ndarray]:

            background_nframes = np.empty(0, dtype=np.int64)
            star_nframes = np.empty(0, dtype=np.int64)

            background_index = np.empty(0, dtype=np.int64)
            star_index = np.empty(0, dtype=np.int64)

            if parang is None:
                background_parang = None
                star_parang = None

            else:
                background_parang = np.empty(0, dtype=np.float64)
                star_parang = np.empty(0, dtype=np.float64)

            return star_index, star_parang, star_nframes, background_index, \
                background_parang, background_nframes

        @typechecked
        def _select_background(i: int) -> np.ndarray:

            # Previous background cube
            if np.size(bg_indices[bg_indices < i]) > 0:
                index_prev = np.amax(bg_indices[bg_indices < i])
                bg_prev = cube_mean[index_prev, ]

            else:
                bg_prev = None

            # Next background cube
            if np.size(bg_indices[bg_indices > i]) > 0:
                index_next = np.amin(bg_indices[bg_indices > i])
                bg_next = cube_mean[index_next, ]

            else:
                bg_next = None

            # Select background: previous, next, or mean of previous and next
            if bg_prev is None and bg_next is not None:
                background = bg_next

            elif bg_prev is not None and bg_next is None:
                background = bg_prev

            elif bg_prev is not None and bg_next is not None:
                background = (bg_prev+bg_next)/2.

            else:
                raise ValueError('Neither previous nor next background frames are found.')

            return background

        nframes = self.m_image_in_port.get_attribute('NFRAMES')
        index = self.m_image_in_port.get_attribute('INDEX')

        star_index, star_parang, star_nframes, background_index, background_parang, \
            background_nframes = _initialize()

        # Separate star and background cubes. Subtract mean background.
        count = 0

        start_time = time.time()
        for i, item in enumerate(nframes):
            progress(i, len(nframes), 'Preparing PCA background subtraction...', start_time)

            im_tmp = self.m_image_in_port[count:count+item, ]

            # Background frames
            if bg_frames[i]:

                self.m_background_out_port.append(im_tmp)

                background_nframes = np.append(background_nframes, nframes[i])
                background_index = np.append(background_index, index[count:count+item])

                if parang is not None:
                    background_parang = np.append(background_parang, parang[count:count+item])

            # Star frames
            else:
                background = _select_background(i)

                self.m_star_out_port.append(im_tmp)
                self.m_subtracted_out_port.append(im_tmp-background)

                star_nframes = np.append(star_nframes, nframes[i])
                star_index = np.append(star_index, index[count:count+item])

                if parang is not None:
                    star_parang = np.append(star_parang, parang[count:count+item])

            count += item

        return star_index, star_parang, star_nframes, background_index, \
            background_parang, background_nframes

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Separates the star and background frames, subtracts the mean
        or median background from both the star and background frames, and writes the star and
        background frames separately to their respective output ports.

        Returns
        -------
        NoneType
            None
        """

        if 'PARANG' in self.m_image_in_port.get_all_non_static_attributes():
            parang = self.m_image_in_port.get_attribute('PARANG')
        else:
            parang = None

        bg_frames, cube_mean = self._prepare()

        bg_indices = np.nonzero(bg_frames)[0]

        star_index, star_parang, star_nframes, background_index, background_parang, \
            background_nframes = self._separate(bg_frames, bg_indices, parang, cube_mean)

        history = f'frames = {sum(star_nframes)}, {len(background_nframes)}'

        self.m_star_out_port.copy_attributes(self.m_image_in_port)
        self.m_star_out_port.add_history('PCABackgroundPreparationModule', history)
        self.m_star_out_port.add_attribute('NFRAMES', star_nframes, static=False)
        self.m_star_out_port.add_attribute('INDEX', star_index, static=False)

        if parang is not None:
            self.m_star_out_port.add_attribute('PARANG', star_parang, static=False)

        self.m_subtracted_out_port.copy_attributes(self.m_image_in_port)
        self.m_subtracted_out_port.add_history('PCABackgroundPreparationModule', history)
        self.m_subtracted_out_port.add_attribute('NFRAMES', star_nframes, static=False)
        self.m_subtracted_out_port.add_attribute('INDEX', star_index, static=False)

        if parang is not None:
            self.m_subtracted_out_port.add_attribute('PARANG', star_parang, static=False)

        self.m_background_out_port.copy_attributes(self.m_image_in_port)
        self.m_background_out_port.add_history('PCABackgroundPreparationModule', history)
        self.m_background_out_port.add_attribute('NFRAMES', background_nframes, static=False)
        self.m_background_out_port.add_attribute('INDEX', background_index, static=False)

        if parang is not None:
            self.m_background_out_port.add_attribute('PARANG', background_parang, static=False)

        self.m_star_out_port.close_port()


class PCABackgroundSubtractionModule(ProcessingModule):
    """
    Pipeline module applying a PCA-based background subtraction (see Hunziker et al. 2018).
    """

    __author__ = 'Tomas Stolker, Silvan Hunziker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 star_in_tag: str,
                 background_in_tag: str,
                 residuals_out_tag: str,
                 fit_out_tag: Optional[str] = None,
                 mask_out_tag: Optional[str] = None,
                 pca_number: int = 60,
                 mask_star: float = 0.7,
                 subframe: Optional[float] = None,
                 gaussian: float = 0.15,
                 **kwargs) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the pipeline module instance.
        star_in_tag : str
            Database tag with the input images that contain the star.
        background_in_tag : str
            Database tag with the input images that contain only background emission.
        residuals_out_tag : str
            Database tag to store the background-subtracted images of the star.
        fit_out_tag : str, None
            Database tag to store the modeled background images. The data is not stored if the
            arguments is set to None.
        mask_out_tag : str, None
            Database tag to store the mask. The data is not stored if the argument is set to None.
        pca_number : int
            Number of principal components that is used to model the background emission.
        mask_star : float
            Radius of the central mask (arcsec).
        gaussian : float
            Full width at half maximum (arcsec) of the Gaussian kernel that is used to smooth the
            image before the star is located.
        subframe : float, None
            Size (arcsec) of the subframe that is used to find the star. Cropping of the subframe
            is done around the center of the image. The full images is used if set to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        if 'subtract_mean' in kwargs:
            warnings.warn('The \'subtract_mean\' parameter has been deprecated. Subtracting of '
                          'the mean is no longer optional so subtract_mean=True.',
                          DeprecationWarning)

        self.m_star_in_port = self.add_input_port(star_in_tag)
        self.m_background_in_port = self.add_input_port(background_in_tag)
        self.m_residuals_out_port = self.add_output_port(residuals_out_tag)

        if fit_out_tag is None:
            self.m_fit_out_port = None
        else:
            self.m_fit_out_port = self.add_output_port(fit_out_tag)

        if mask_out_tag is None:
            self.m_mask_out_port = None
        else:
            self.m_mask_out_port = self.add_output_port(mask_out_tag)

        self.m_pca_number = pca_number
        self.m_mask_star = mask_star
        self.m_gaussian = gaussian
        self.m_subframe = subframe

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Creates a PCA basis set of the background frames after
        subtracting the mean background frame from both the star and background frames, masks the
        PSF of the star, projects the star frames onto the principal components, and stores the
        residuals of the background subtracted images.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _create_mask(radius: float,
                         position: np.ndarray,
                         nimages: int) -> np.ndarray:
            """
            Method for creating a circular mask at the star or planet position.
            """

            npix = self.m_star_in_port.get_shape()[1]

            x_grid = np.arange(0, npix, 1)
            y_grid = np.arange(0, npix, 1)

            xx_grid, yy_grid = np.meshgrid(x_grid, y_grid)

            mask = np.ones((nimages, npix, npix))

            cent_x = position[:, 1]
            cent_y = position[:, 0]

            for i in range(nimages):
                rr_grid = np.sqrt((xx_grid - cent_x[i])**2 + (yy_grid - cent_y[i])**2)
                mask[i, ][rr_grid < radius] = 0.

            return mask

        @typechecked
        def _create_basis(images: np.ndarray,
                          pca_number: int) -> np.ndarray:
            """
            Method for calculating the principal components for a stack of background images.

            Parameters
            ----------
            images : np.ndarray
                Background images with the mean subtracted from all images.
            pca_number : int
                Number of principal components that is used to model the background emission.

            Returns
            -------
            np.ndarray
                Principal components with the second and third dimension reshaped to ``images``.
            """

            _, _, v_svd = svds(images.reshape(images.shape[0],
                                              images.shape[1]*images.shape[2]),
                               k=pca_number)

            v_svd = v_svd[::-1, ]

            return v_svd.reshape(v_svd.shape[0], images.shape[1], images.shape[2])

        @typechecked
        def _model_background(basis: np.ndarray,
                              im_arr: np.ndarray,
                              mask: np.ndarray) -> np.ndarray:
            """
            Method for creating a model of the background.
            """

            @typechecked
            def _dot_product(x_dot: np.ndarray,
                             *p: np.float64) -> np.ndarray:
                return np.dot(p, x_dot)

            fit_im_chi = np.zeros(im_arr.shape)
            # fit_coeff_chi = np.zeros((im_arr.shape[0], basis.shape[0]))

            basis_reshaped = basis.reshape(basis.shape[0], -1)

            for i in range(im_arr.shape[0]):
                basis_reshaped_masked = (basis*mask[i]).reshape(basis.shape[0], -1)

                data_to_fit = im_arr[i, ]

                init = np.ones(basis_reshaped_masked.shape[0])

                fitted = curve_fit(_dot_product,
                                   basis_reshaped_masked,
                                   data_to_fit.reshape(-1),
                                   init)

                fit_im = np.dot(fitted[0], basis_reshaped)
                fit_im = fit_im.reshape(data_to_fit.shape[0], data_to_fit.shape[1])

                fit_im_chi[i, ] = fit_im
                # fit_coeff_chi[i, ] = fitted[0]

            return fit_im_chi

        memory = self._m_config_port.get_attribute('MEMORY')
        pixscale = self.m_star_in_port.get_attribute('PIXSCALE')

        nimages = self.m_star_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        self.m_mask_star /= pixscale
        self.m_gaussian = int(math.ceil(self.m_gaussian/pixscale))

        if self.m_subframe is not None:
            self.m_subframe /= pixscale
            self.m_subframe = int(math.ceil(self.m_subframe))

        bg_mean = np.mean(self.m_background_in_port.get_all(), axis=0)

        star = np.zeros((nimages, 2))
        for i, _ in enumerate(star):
            star[i, :] = locate_star(image=self.m_star_in_port[i, ] - bg_mean,
                                     center=None,
                                     width=self.m_subframe,
                                     fwhm=self.m_gaussian)

        print('Creating PCA basis set...', end='')

        basis_pca = _create_basis(self.m_background_in_port.get_all() - bg_mean,
                                  self.m_pca_number)

        print(' [DONE]')

        start_time = time.time()
        for i, _ in enumerate(frames[:-1]):
            progress(i, len(frames[:-1]), 'Calculating background model...', start_time)

            # Subtract the mean background from the star frames
            im_star = self.m_star_in_port[frames[i]:frames[i+1], ] - bg_mean

            mask = _create_mask(self.m_mask_star,
                                star[frames[i]:frames[i+1], ],
                                int(frames[i+1]-frames[i]))

            fit_im = _model_background(basis_pca, im_star*mask, mask)

            self.m_residuals_out_port.append(im_star-fit_im)

            if self.m_fit_out_port is not None:
                self.m_fit_out_port.append(fit_im)

            if self.m_mask_out_port is not None:
                self.m_mask_out_port.append(mask)

        history = f'PC number = {self.m_pca_number}'
        self.m_residuals_out_port.copy_attributes(self.m_star_in_port)
        self.m_residuals_out_port.add_history('PCABackgroundSubtractionModule', history)
        self.m_residuals_out_port.add_attribute('STAR_POSITION', star, static=False)

        if self.m_fit_out_port is not None:
            self.m_fit_out_port.copy_attributes(self.m_star_in_port)
            self.m_fit_out_port.add_history('PCABackgroundSubtractionModule', history)

        if self.m_mask_out_port is not None:
            self.m_mask_out_port.copy_attributes(self.m_star_in_port)
            self.m_mask_out_port.add_history('PCABackgroundSubtractionModule', history)

        self.m_residuals_out_port.close_port()


class DitheringBackgroundModule(ProcessingModule):
    """
    Pipeline module for PCA-based background subtraction of dithering data. This is a wrapper that
    applies the processing modules for either a mean or the PCA-based background subtraction.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 image_out_tag: str,
                 center: Optional[List[Tuple[int, int]]] = None,
                 cubes: Optional[int] = None,
                 size: float = 2.,
                 gaussian: float = 0.15,
                 subframe: Optional[float] = None,
                 pca_number: Optional[int] = 5,
                 mask_star: float = 0.7,
                 **kwargs) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Database tag with input images.
        image_out_tag : str
            Database tag to store the background subtracted images.
        center : list(tuple(int, int)), None
            Tuple with the centers of the dithering positions, e.g. ((x0, y0), (x1, y1)). The order
            of the coordinates should correspond to the order in which the star is present. If
            ``center`` and ``cubes`` are both set to None then sorting and subtracting of the
            background frames is based on ``DITHER_X`` and ``DITHER_Y``. If ``center`` is
            specified and ``cubes`` is set to None then the ``DITHER_X`` and ``DITHER_Y``
            attributes will be used for sorting and subtracting of the background but not for
            selecting the dither positions.
        cubes : int, None
            Number of consecutive cubes per dither position. If ``cubes`` is set to None then
            sorting and subtracting of the background frames is based on ``DITHER_X`` and
            ``DITHER_Y``.
        size : float
            Cropped image size (arcsec).
        gaussian : float
            Full width at half maximum (arcsec) of the Gaussian kernel that is used to smooth the
            image before the star is located.
        subframe : float, None
            Size (arcsec) of the subframe that is used to search for the star. Cropping of the
            subframe is done around the center of the dithering position. The full image size
            (i.e. ``size``) will be used if set to None then.
        pca_number : int, None
            Number of principal components that is used to model the background emission. The PCA
            background subtraction is skipped if the argument is set to None. In that case, the
            mean background subtracted images are written toe ``image_out_tag``.
        mask_star : float
            Radius of the central mask (arcsec) that is used to exclude the star when fitting the
            principal components. The region behind the mask is included when subtracting the
            PCA background model.

        Returns
        -------
        NoneType
            None
        """

        if 'mask_planet' in kwargs:
            warnings.warn('The \'mask_planet\' parameter has been deprecated.', DeprecationWarning)

        if 'crop' in kwargs:
            warnings.warn('The \'crop\' parameter has been deprecated. The step to crop the '
                          'images is no longer optional so crop=True.', DeprecationWarning)

        if 'prepare' in kwargs:
            warnings.warn('The \'prepare\' parameter has been deprecated. The preparation step '
                          'is no longer optional so prepare=True.', DeprecationWarning)

        if 'pca_background' in kwargs:
            warnings.warn('The \'pca_background\' parameter has been deprecated. The PCA '
                          'background is no longer optional when combine=\'pca\' so '
                          'pca_background=True.', DeprecationWarning)

        if 'subtract_mean' in kwargs:
            warnings.warn('The \'subtract_mean\' parameter has been deprecated. Subtracting of '
                          'the mean is no longer optional so subtract_mean=True.',
                          DeprecationWarning)

        if 'combine' in kwargs:
            warnings.warn('The \'combine\' parameter has been deprecated. To write the mean '
                          'background subtracted images to image_out_tag is done by setting '
                          'pca_number=None.', DeprecationWarning)

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_center = center
        self.m_cubes = cubes
        self.m_size = size
        self.m_gaussian = gaussian
        self.m_pca_number = pca_number
        self.m_mask_star = mask_star
        self.m_subframe = subframe

        self.m_image_in_tag = image_in_tag
        self.m_image_out_tag = image_out_tag

    @typechecked
    def _initialize(self) -> Tuple[int, np.ndarray]:
        if self.m_cubes is None:
            dither_x = self.m_image_in_port.get_attribute('DITHER_X')
            dither_y = self.m_image_in_port.get_attribute('DITHER_Y')

            dither_xy = np.zeros((dither_x.shape[0], 2))
            dither_xy[:, 0] = dither_x
            dither_xy[:, 1] = dither_y

            _, index = np.unique(dither_xy, axis=0, return_index=True)

            dither = dither_xy[np.sort(index)]

            npix = self.m_image_in_port.get_shape()[1]

            # Compute center from dither and make sure all positions are actually Python integers
            if self.m_center is None:
                self.m_center = np.copy(dither) + float(npix) / 2.

                self.m_center = tuple(zip(map(int, self.m_center[:, 0]),
                                          map(int, self.m_center[:, 1])))

            else:
                if np.size(dither, axis=0) != np.size(self.m_center, axis=0):
                    raise ValueError('Number of specified center positions should be equal to the '
                                     'number of unique dithering positions.')

        n_dither = np.size(self.m_center, 0)

        if self.m_cubes is None:
            star_pos = np.copy(dither)
        else:
            star_pos = np.arange(0, n_dither, 1)

        return n_dither, star_pos

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Cuts out the detector sections at the different dither positions,
        prepares the PCA background subtraction, locates the star in each image, runs the PCA
        background subtraction, combines the output from the different dither positions is written
        to a single database tag.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _admin_start(count: int,
                         n_dither: int,
                         position: Tuple[int, int],
                         star_pos: Union[np.ndarray, np.int64]) -> None:
            print(f'Processing dither position {count+1} out of {n_dither}...')
            print(f'Center position = {position}')

            if self.m_cubes is None and self.m_center is not None:
                print(f'DITHER_X, DITHER_Y = {tuple(star_pos)}')

        @typechecked
        def _admin_end(count: int) -> None:
            if self.m_pca_number is None:
                tags.append(f'{self.m_image_in_tag}_dither_mean{count+1}')

            else:
                tags.append(f'{self.m_image_in_tag}_dither_pca_res{count+1}')

        n_dither, star_pos = self._initialize()
        tags = []

        for i, position in enumerate(self.m_center):
            _admin_start(i, n_dither, position, star_pos[i])

            im_out_tag = f'{self.m_image_in_tag}_dither_crop{i+1}'

            module = CropImagesModule(name_in=f'crop{i}',
                                      image_in_tag=self.m_image_in_tag,
                                      image_out_tag=im_out_tag,
                                      size=self.m_size,
                                      center=(int(math.ceil(position[0])),
                                              int(math.ceil(position[1]))))

            module.connect_database(self._m_data_base)
            module._m_output_ports[im_out_tag].del_all_data()
            module._m_output_ports[im_out_tag].del_all_attributes()
            module.run()

            if self.m_cubes is None:
                dither_val = (n_dither, self.m_cubes, tuple(star_pos[i]))
            else:
                dither_val = (n_dither, self.m_cubes, int(star_pos[i]))

            im_in_tag = f'{self.m_image_in_tag}_dither_crop{i+1}'
            star_out_tag = f'{self.m_image_in_tag}_dither_star{i+1}'
            sub_out_tag = f'{self.m_image_in_tag}_dither_mean{i+1}'
            back_out_tag = f'{self.m_image_in_tag}_dither_background{i+1}'

            module = PCABackgroundPreparationModule(name_in=f'prepare{i}',
                                                    image_in_tag=im_in_tag,
                                                    star_out_tag=star_out_tag,
                                                    subtracted_out_tag=sub_out_tag,
                                                    background_out_tag=back_out_tag,
                                                    dither=dither_val,
                                                    combine='mean')

            module.connect_database(self._m_data_base)
            module._m_output_ports[star_out_tag].del_all_data()
            module._m_output_ports[star_out_tag].del_all_attributes()
            module._m_output_ports[sub_out_tag].del_all_data()
            module._m_output_ports[sub_out_tag].del_all_attributes()
            module._m_output_ports[back_out_tag].del_all_data()
            module._m_output_ports[back_out_tag].del_all_attributes()
            module.run()

            if self.m_pca_number is not None:
                star_in_tag = f'{self.m_image_in_tag}_dither_star{i+1}'
                back_in_tag = f'{self.m_image_in_tag}_dither_background{i+1}'
                res_out_tag = f'{self.m_image_in_tag}_dither_pca_res{i+1}'
                fit_out_tag = f'{self.m_image_in_tag}_dither_pca_fit{i+1}'
                mask_out_tag = f'{self.m_image_in_tag}_dither_pca_mask{i+1}'

                module = PCABackgroundSubtractionModule(name_in=f'pca_background{i}',
                                                        star_in_tag=star_in_tag,
                                                        background_in_tag=back_in_tag,
                                                        residuals_out_tag=res_out_tag,
                                                        fit_out_tag=fit_out_tag,
                                                        mask_out_tag=mask_out_tag,
                                                        pca_number=self.m_pca_number,
                                                        mask_star=self.m_mask_star,
                                                        subframe=self.m_subframe,
                                                        gaussian=self.m_gaussian)

                module.connect_database(self._m_data_base)
                module._m_output_ports[res_out_tag].del_all_data()
                module._m_output_ports[res_out_tag].del_all_attributes()
                module._m_output_ports[fit_out_tag].del_all_data()
                module._m_output_ports[fit_out_tag].del_all_attributes()
                module._m_output_ports[mask_out_tag].del_all_data()
                module._m_output_ports[mask_out_tag].del_all_attributes()
                module.run()

            _admin_end(i)

        module = CombineTagsModule(name_in='combine',
                                   check_attr=True,
                                   index_init=False,
                                   image_in_tags=tags,
                                   image_out_tag=self.m_image_in_tag+'_dither_combine')

        module.connect_database(self._m_data_base)
        module._m_output_ports[self.m_image_in_tag+'_dither_combine'].del_all_data()
        module._m_output_ports[self.m_image_in_tag+'_dither_combine'].del_all_attributes()
        module.run()

        module = SortParangModule(name_in='sort',
                                  image_in_tag=self.m_image_in_tag+'_dither_combine',
                                  image_out_tag=self.m_image_out_tag)

        module.connect_database(self._m_data_base)
        module._m_output_ports[self.m_image_out_tag].del_all_data()
        module._m_output_ports[self.m_image_out_tag].del_all_attributes()
        module.run()
